/*
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * cache_attrib.c
 *
 * Author: 	Renato J. Figueiredo
 *		Ming Zhao
 *
 * 		Functions that implement the gvfs proxy attribute cache.
 * 		The attribute cache is stored in the local disk.
 * 		Refer to cache_attrib.h for a description of the attribute cache
 * 		organization
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <unistd.h>
#include <fcntl.h>
#include <math.h>

#include "nfs_prot.h"
#include "logging.h"
#include "cache_attrib.h"

/* 
 * Global variables
 */

unsigned long int cache_size = MAX_CACHE_SIZE;
unsigned int cache_assoc = MAX_CACHE_ASSOC;
unsigned int cache_banks = MAX_CACHE_BANKS;
unsigned long int cache_sets; /* total number of cache sets */
unsigned long int sets_per_bank;
char *BaseCachePath;
char *SessionID;
char *scratch_path;
struct stat *statbuf;
cacheblock *block_scratch_RPC;
cacheblock *block_scratch_disk;

int flush_attr_cache(void) {
  int staterr;
  unsigned long int i;

  Dprintf(D_ACACHE, "Flushing attribute cache...\n");
  for(i=0; i<cache_banks; i++) {
    /* Not yet checking for modified blocks!!! */
    if((strlen(BaseCachePath)+1+strlen(SessionID)+
	1+strlen("attr_cache_bank_")+16)>=MAX_STR) {
      Dprintf(L_ERROR,"cache block path too long\n");
      return -1;
    }
    sprintf(scratch_path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,
	    SessionID,i);
    staterr = stat(scratch_path, statbuf);
    if(staterr==0) { /* file exists; remove it */
      if((unlink(scratch_path))==-1) {
	Dprintf(L_ERROR, "Could not flush attribute cache bank!\n");
	return -1;
      }
    }
  }
  Dprintf(D_ACACHE, "Done flushing attribute cache.\n");
  return 0;
}

/*************************************************************************
 * NFS-V2-LOOKUP cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 lookup arguments into cacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int argument_lookup_v2(diropargs *arg, cacheblock *block) {

  if(arg==NULL) {
    Dprintf(D_ACACHE,"argument_lookup_v2: arg==NULL\n");
    return -1;
  }

  /* zero out cache block */
  bzero(block,sizeof(*block));

  /* mark block as NFS V2 lookup */
  block->nfsprogid = 2;
  block->nfsprocid = NFSPROC_LOOKUP;

  /* copy contents of argument to cache block */
  if((memcpy(&block->contents.lookup_v2.dir,&arg->dir,NFS_FHSIZE))==NULL) {
    Dprintf(D_ACACHE,"argument_lookup_v2: memcpy failed\n");
    return -1;
  }
  if(strlen(arg->name) > MAX_STR) /* make sure string fits */ {
    Dprintf(D_ACACHE,"argument_lookup_v2: string larger than MAX_STR\n");    
    return -1;
  }
  strcpy((char *)&block->contents.lookup_v2.name,arg->name);
  return 0;
}

/*************************************************************************
 * Copy NFS-V2 lookup results into cacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_remote_RPC_lookup_v2(diropres *res, cacheblock *block) {

  if(res==NULL) {
    Dprintf(D_ACACHE,"response_remote_RPC_lookup_v2: res==NULL\n");
    return -1;
  }

  block->contents.lookup_v2.status = res->status;
  if((memcpy(&block->contents.lookup_v2.file,
	     &res->diropres_u.diropres.file,NFS_FHSIZE))==NULL) {
    Dprintf(D_ACACHE,"response_remote_RPC_lookup_v2: memcpy failed\n");
    return -1;
  }
  block->contents.lookup_v2.attributes = res->diropres_u.diropres.attributes;
  return 0;
}

/*************************************************************************
 * Copy NFS-V2 lookup cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_from_disk_cache_lookup_v2(cacheblock *block, diropres *res) {

  if(res==NULL) {
    Dprintf(D_ACACHE,"response_from_disk_cache_lookup_v2: res==NULL\n");
    return -1;
  }
  res->status = block->contents.lookup_v2.status;
  if((memcpy(&res->diropres_u.diropres.file,
	     &block->contents.lookup_v2.file,
	     NFS_FHSIZE))==NULL) {
    Dprintf(D_ACACHE,"response_from_disk_cache_lookup_v2: memcpy failed\n");
    return -1;
  }
  res->diropres_u.diropres.attributes = block->contents.lookup_v2.attributes;
  return 0;
}

/*************************************************************************
 * Hash V2 diropargs (fh of directory + file name)                       *
 * Return set number                                                     *
 * HERE: check efficiency of this hash                                   *
 *************************************************************************/
unsigned long int hash_diropargs_v2(nfs_fh *fhandle, char *filename) {
  int i,seof=0;
  double A=0.6180339887;
  unsigned long int B=0;

  /* accumulate file handle into B */
  for(i=0;i<NFS_FHSIZE;i++) {
    B+=(unsigned long int)(fhandle->data[i]);
  }

  /* also accumulate file name */
  for(i=0;seof==0;i++) {
    if(filename[i] == 0)
      seof=1;
    else
      B+=(unsigned long int)(filename[i]);
  }

  A = A * (double)B;
  A = A - floor(A);
  A = A * cache_sets;
  return (unsigned long int) (A) % cache_sets;
}

/*************************************************************************
 * NFS-V2-GETATTR cache entry manipulation functions
 *************************************************************************/

/*************************************************************************
 * Copy NFS-V2 getattr arguments into cacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int argument_getattr_v2(nfs_fh *arg, cacheblock *block) {

  if(arg==NULL) {
    Dprintf(D_ACACHE,"argument_getattr_v2: arg==NULL\n");
    return -1;
  }

  /* zero out cache block */
  bzero(block,sizeof(*block));

  /* mark block as NFS V2 getattr */
  block->nfsprogid = 2;
  block->nfsprocid = NFSPROC_GETATTR;

  /* copy contents of argument to cache block */
  if((memcpy(&block->contents.getattr_v2.handle,
	     arg,NFS_FHSIZE))==NULL) {
    Dprintf(D_ACACHE,"argument_getattr_v2: memcpy failed\n");
    return -1;
  }
  return 0;
}

/*************************************************************************
 * Copy NFS-V2 getattr results into cacheblock data structure 
 * that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_remote_RPC_getattr_v2(attrstat *res, cacheblock *block) {

  if(res==NULL) {
    Dprintf(D_ACACHE,"response_remote_RPC_getattr_v2: res==NULL\n");
    return -1;
  }

  block->contents.getattr_v2.status = res->status;
  block->contents.getattr_v2.attributes = 
    res->attrstat_u.attributes;

  return 0;
}

/*************************************************************************
 * Copy NFS-V2 getattr cached data from cacheblock data structure to RPC
 * response that is passed as argument. Returns 0 if successful, -1 if
 * error                             
 *************************************************************************/
int response_from_disk_cache_getattr_v2(cacheblock *block, attrstat *res) {

  if(res==NULL) {
    Dprintf(D_ACACHE,"response_from_disk_cache_getattr_v2: res==NULL\n");
    return -1;
  }
  res->status = block->contents.getattr_v2.status;
  res->attrstat_u.attributes = block->contents.getattr_v2.attributes;
  return 0;
}

unsigned long int hash_nfs_fh_v2(nfs_fh *fhandle) {
  return hash_diropargs_v2(fhandle,"xx");
}

/*************************************************************************
 * Determine the file name that a cache block maps to                    *
 * Returns set number,                                                   *
 *  and the path to the cache bank file if successfull, -1 otherwise     *
 *************************************************************************/
long int block_file_path_and_set(cacheblock *block, char *path) {
  unsigned long int hash, bank_number;

  if(block==NULL) {
    Dprintf(D_ACACHE,"block_file_path_and_set: block==NULL\n");
    return -1;
  }

  switch(block->nfsprogid) {
  case 2:
    switch(block->nfsprocid) {
    case NFSPROC_LOOKUP:
      hash = hash_diropargs_v2(&(block->contents.lookup_v2.dir),
			       (char *)&(block->contents.lookup_v2.name));
      bank_number = (unsigned long int) (hash/sets_per_bank);
      if((strlen(BaseCachePath)+1+strlen(SessionID)+
	  1+strlen("attr_cache_bank_")+16)>MAX_STR) {
	Dprintf(L_ERROR,"cache block path too long\n");
	return -1;
      }
      sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
	      bank_number);
      Dprintf(D_ACACHE,"HASH lookup = %lu (bank %lu)\n",hash,bank_number);
      return (unsigned long int) hash;
      break;
    case NFSPROC_GETATTR:
      hash = hash_nfs_fh_v2(&(block->contents.getattr_v2.handle));
      bank_number = (unsigned long int) (hash/sets_per_bank);
      if((strlen(BaseCachePath)+1+strlen(SessionID)+
	  1+strlen("attr_cache_bank_")+16)>MAX_STR) {
	Dprintf(L_ERROR,"cache block path too long\n");
	return -1;
      }
      sprintf(path,"%s/%s/attr_cache_bank_%lu",BaseCachePath,SessionID,
	      bank_number);
      Dprintf(D_ACACHE,"HASH(%d) = %lu (bank %lu)\n",
      		block->nfsprocid,hash,bank_number);
      return (unsigned long int) hash;
    default:
      return -1;
    }
    break;
  default:
    return -1;
  }
  return -1;
}

/*************************************************************************
 * Check whether tag of NFS arguments matches cached data on disk,       * 
 *  possibly invalidating cached data (if invalidate==1)                 *
 * block_RPC : has NFS arguments                                         *
 * block_disk: returns contents of a match on a hit                      *
 * return values:                                                        *
 * 0 : no match                                                          *
 * 1 : tag match                                                         *
 * -1: error                                                             *
 *************************************************************************/
int tag_match_and_inv(cacheblock *block_RPC, cacheblock *block_disk,
		      int invalidate) {
  int fd,reads,framescan;
  int file_handle_match = 0;
  int file_name_match = 0;
  int overall_match = 0;
  int seekset, seekr, seekcur;
  unsigned long int set_number;
  int staterr;

  if(block_RPC==NULL) {
    Dprintf(D_ACACHE,"tag_match: block_RPC==NULL\n");
    return -1;
  }

  /* obtain path and set number */
  set_number = block_file_path_and_set(block_RPC, scratch_path);
  if(set_number==-1) {
    Dprintf(D_ACACHE,"tag_match: could not obtain set\n");
    return -1;
  }

  /* open & seek into 1st block of appropriate set in cache file */
  staterr = stat(scratch_path, statbuf);
  if(staterr != 0) {
    Dprintf(D_ACACHE, "tag_match: bank does not exist\n");
    return 0;
  }
  if((fd = open(scratch_path, O_RDONLY)) == -1) {
    Dprintf(D_ACACHE,"tag_match: could not open cache bank %s for reading\n",
	    scratch_path);
    return -1;
  }

  /* obtain a shared lock for cache lookup */
  if ((flock(fd,LOCK_SH))==-1) {
    Dprintf(D_ACACHE,"tag_match: could not lock cache bank %s for reading\n",
	    scratch_path);
    return -1;
  }

  seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(*block_RPC);
  Dprintf(D_ACACHE, "tag_match: seeking %lu -> %lu\n",set_number,seekset);
  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_ACACHE,"tag_match: could not seek initial block %i in %s\n",
	    seekset, scratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* associative lookup; traverse all sets */
  for(framescan=0; framescan<cache_assoc && !overall_match; framescan++) {

    reads = read(fd,block_disk,sizeof(*block_disk));
    if(reads != sizeof(*block_disk)) { /* read was not right */
      Dprintf(D_ACACHE,"tag_match: could not read scan block %i in %s\n",
	      framescan, scratch_path);
      flock(fd,LOCK_UN);
      close(fd);
      return -1;
    }

    /* different NFS calls require diff. tag matching */
    switch(block_RPC->nfsprogid) {
    case 2:
      switch(block_RPC->nfsprocid) {
      case NFSPROC_LOOKUP:
	if (block_disk->state == ST_INV) /* invalid block; no match */
	  break;
	/* check if blocks' prog/proc ids match */
	if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
	    (block_RPC->nfsprocid==block_disk->nfsprocid))) 
	  break;
	/* check if file names match */
	file_name_match = strcmp((char *)&block_RPC->contents.lookup_v2.name,
				 (char *)&block_disk->contents.lookup_v2.name);
	if (file_name_match != 0)  /* no file name match */
	  break;
	/* check if file handles match */
	file_handle_match = memcmp(&block_RPC->contents.lookup_v2.dir,
				   &block_disk->contents.lookup_v2.dir,
				   NFS_FHSIZE);
	overall_match = (file_handle_match==0? 1 : 0);
	if(!overall_match) break;
	
	/* invalidate block in disk if required */
	if(invalidate) {
	  /* keep current seek point in seekcur; will need it after
	     fd is reopened */
	  seekcur = lseek(fd, 0, SEEK_CUR);
	  /* release shared lock  */
	  if ((flock(fd,LOCK_UN))==-1) {
	    Dprintf(D_ACACHE,
		    "tag_match: could not unlock cache bank %s for reading\n",
		    scratch_path);
	    return -1;
	  }
	  /* reopen bank file in read/write mode  for invalidation */
	  close(fd);
	  if((fd = open(scratch_path, O_RDWR)) == -1) {
	    Dprintf(D_ACACHE,"tag_match: could not open cache bank %s R/W\n",
		    scratch_path);
	    return -1;
	  }
	  /* obtain exclusive lock for invalidation */
	  if ((flock(fd,LOCK_EX))==-1) {
	    Dprintf(D_ACACHE,
		    "tag_match: could not lock cache bank %s for writing\n",
		    scratch_path);
	    return -1;
	  }
	  /* seek back one block from seekcur */
	  seekset = seekcur-sizeof(*block_RPC); /* point back one block */
	  seekr = lseek(fd, seekset, SEEK_SET);
	  if(seekr!=seekset) {
	    Dprintf(D_ACACHE, "tag_match: could not seek %i in file bank %s\n",
		    seekset, scratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }
	  /* write an invalid block (all zeroes) */
	  bzero(block_disk,sizeof(*block_disk));
	  if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
	    Dprintf(D_ACACHE,"tag_match: could not invalidate block in %s\n",
		    scratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }	  
	}
	break;
      case NFSPROC_GETATTR:
	if (block_disk->state == ST_INV) /* invalid block; no match */
	  break;
	/* check if blocks' prog/proc ids match */
	if(!((block_RPC->nfsprogid==block_disk->nfsprogid) &&
	    (block_RPC->nfsprocid==block_disk->nfsprocid)))
	  break;
	/* check if file handles match */
	file_handle_match = memcmp(&block_RPC->contents.getattr_v2.handle,
				   &block_disk->contents.getattr_v2.handle,
				   NFS_FHSIZE);
	overall_match = (file_handle_match==0 ? 1 : 0);
	if(!overall_match) break;
	
	/* invalidate block in disk if required */
	if(invalidate) {
	  /* keep current seek point in seekcur; will need it after
	     fd is reopened */
	  seekcur = lseek(fd, 0, SEEK_CUR);
	  /* release shared lock  */
	  if ((flock(fd,LOCK_UN))==-1) {
	    Dprintf(D_ACACHE,
		    "tag_match: could not unlock cache bank %s for reading\n",
		    scratch_path);
	    return -1;
	  }
	  /* reopen bank file in read/write mode  for invalidation */
	  close(fd);
	  if((fd = open(scratch_path, O_RDWR)) == -1) {
	    Dprintf(D_ACACHE,"tag_match: could not open cache bank %s R/W\n",
		    scratch_path);
	    return -1;
	  }
	  /* obtain exclusive lock for invalidation */
	  if ((flock(fd,LOCK_EX))==-1) {
	    Dprintf(D_ACACHE,
		    "tag_match: could not lock cache bank %s for writing\n",
		    scratch_path);
	    return -1;
	  }
	  /* seek back one block from seekcur */
	  seekset = seekcur-sizeof(*block_RPC); /* point back one block */
	  seekr = lseek(fd, seekset, SEEK_SET);
	  if(seekr!=seekset) {
	    Dprintf(D_ACACHE, "tag_match: could not seek %i in file bank %s\n",
		    seekset, scratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }
	  /* write an invalid block (all zeroes) */
	  bzero(block_disk,sizeof(*block_disk));
	  if((write(fd,block_disk,sizeof(*block_disk)))!=sizeof(*block_disk)) {
	    Dprintf(D_ACACHE,"tag_match: could not invalidate block in %s\n",
		    scratch_path);
	    flock(fd,LOCK_UN);
	    close(fd);
	    return -1;
	  }	  
	} /* end if invalidate */
	break;
      }
      break;
    }
  }
  /* release shared lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_ACACHE,
	    "tag_match: could not unlock cache bank %s for reading\n",
	    scratch_path);
    return -1;
  }
  close(fd);

  Dprintf(D_ACACHE,"tag_match(%d): returns %i (frame %i, state %d)\n",
	    block_RPC->nfsprocid, overall_match, framescan-1, block_disk->state);
  return overall_match;
}

/*************************************************************************
 * Finds if there is a frame with specified state for a cache block      *
 * creates new file for cache bank if necessary                          *
 * return values:                                                        *
 * 0 =< x < cache_assoc        : free frame found in x                   *
 * cache_assoc:                : no free frames found                    *
 * -1: error                                                             *
 *************************************************************************/
int find_replace_frame(unsigned long int set_number, cacheblock *block_disk, 
		       int state) {
  int fd,reads,framescan;
  int overall_match = 0;
  int seekset, seekr;

  /* open & seek into 1st block of appropriate set in cache file */
  if((fd = open(scratch_path, O_RDONLY)) == -1) {
    Dprintf(D_ACACHE,
	    "find_replace_frame: could not open cache bank %s for reading\n",
	    scratch_path);
    return -1;
  }
  /* obtain a shared lock for cache lookup */
  if ((flock(fd,LOCK_SH))==-1) {
    Dprintf(D_ACACHE,
	    "find_replace_frame: could not lock cache bank %s for reading\n",
	    scratch_path);
    return -1;
  }

  seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(cacheblock);
  Dprintf(D_ACACHE,
	  "find_replace_frame: seeking %lu -> %lu\n",set_number,seekset);
  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_ACACHE,
	    "find_replace_frame: could not seek initial block %i in %s\n",
	    seekset, scratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* associative lookup; traverse all sets */
  for(framescan=0; framescan<cache_assoc && !overall_match; framescan++) {

    reads = read(fd,block_disk,sizeof(*block_disk));
    if(reads != sizeof(*block_disk)) { /* read was not right */
      Dprintf(D_ACACHE,
	      "find_replace_frame: could not read scan block %i in %s\n",
	      framescan, scratch_path);
      flock(fd,LOCK_UN);
      close(fd);
      return -1;
    }

    if(block_disk->state == state) {
      /* release shared lock  */
      if ((flock(fd,LOCK_UN))==-1) {
	Dprintf(D_ACACHE,
		"find_replace_frame: could not unlock cache bank %s\n",
		scratch_path);
	return -1;
      }
      close(fd);
      Dprintf(D_ACACHE,
	      "find_replace_frame: found %i(state %d)\n",framescan,block_disk->state);
      return framescan;
    }
  }

  /* release shared lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_ACACHE,
	    "find_replace_frame: could not unlock cache bank %s\n",
	    scratch_path);
    return -1;
  }
  close(fd);
  return cache_assoc;
}


/*************************************************************************
 * Insert result from RPC call in disk cache                             *
 * 0 : inserted                                                          *
 *-1 : error                                                             *
 *************************************************************************/
int insert_in_cache(cacheblock *block_RPC, cacheblock *block_disk) {
  int freeset,replaceset,foundset;
  unsigned long int set_number;
  int staterr, fd;
  int written, seekset, seekr, seekcur;
  char c = '\0';

  if(block_RPC==NULL)
    return -1;

  /* determine if cache bank exists; if not, create it */
  set_number = block_file_path_and_set(block_RPC, scratch_path);
  staterr = stat(scratch_path, statbuf);

  if(staterr!=0) { /* create new bank */
    if((fd = creat(scratch_path, S_IRUSR | S_IWUSR |
		   S_IRGRP | S_IWGRP ))==-1) {
      Dprintf(D_ACACHE,"insert_in_cache: could not create %s for R/W\n",
	      scratch_path);
      return -1;
    }
    lseek(fd, sets_per_bank*cache_assoc*sizeof(*block_disk)-1, SEEK_SET);
    if(write(fd, &c, 1) != 1) {
      Dprintf(D_ACACHE,"insert_in_cache: could not write to %s\n",
	      scratch_path);
      close(fd);
      return -1;
    }
    close(fd);
  } else { /* bank file exists; check permissions */
    if(!((statbuf->st_mode & S_IRUSR) &&
	 (statbuf->st_mode & S_IWUSR))) {
      Dprintf(D_ACACHE,"insert_in_cache: wrong permissions in %s\n",
	      scratch_path);
      return -1;
    }
  }

  /* look for empty block to replace with */
  if((freeset = find_replace_frame(set_number, block_disk, ST_INV))==-1) {
    Dprintf(D_ACACHE,"insert_in_cache: error finding replace frame ST_INV\n");
    return -1;
  }
  if(freeset == cache_assoc) {
    Dprintf(D_ACACHE, "No empty blocks\n");
    /* look for clean block to replace with */
    if((replaceset = find_replace_frame(set_number, block_disk, 
					ST_RO_CLEAN))==-1) {
      Dprintf(D_ACACHE,
	      "insert_in_cache: error finding replace frame ST_RO_CLEAN\n");
      return -1;
    }
    if(replaceset == cache_assoc) {
      Dprintf(D_ACACHE,
	      "insert_in_cache: could not find block to evict\n");
      return -1;
    }
    foundset = replaceset;
    Dprintf(D_ACACHE,"insert_in_cache: replaces block %i\n", replaceset);
  } else {
    foundset = freeset;
    Dprintf(D_ACACHE,"insert_in_cache: found empty frame %i\n", freeset);
  }

  /* open cache bank */
  if((fd = open(scratch_path, O_RDWR))==-1) {
    Dprintf(D_ACACHE, "insert_in_cache: could not open file bank %s\n",
	    scratch_path);
    return -1;
  }

  /* obtain exclusive lock for insertion */
  if ((flock(fd,LOCK_EX))==-1) {
    Dprintf(D_ACACHE,
	    "insert_in_cache: could not lock cache bank %s for writing\n",
	    scratch_path);
    return -1;
  }

  /* seek into appropriate insert position */
  /* first, see into first block of set */
  seekset = (set_number%sets_per_bank)*cache_assoc*sizeof(*block_RPC);
  seekr = lseek(fd,seekset,SEEK_SET);
  if(seekr != seekset) {
    Dprintf(D_ACACHE,
	    "insert in cache: could not seek initial block %i in %s\n",
	    seekset, scratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* then, seek within set for replaced frame */
  seekcur = seekr; /* remember current seek point */
  seekset = foundset*sizeof(*block_RPC); /* determine offset within set */
  seekr = lseek(fd, seekset, SEEK_CUR); /* offset seek */
  if((seekcur+seekset) != seekr) {
    Dprintf(D_ACACHE, "insert_in_cache: could not seek %i in file bank %s\n",
	    seekset, scratch_path);
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }

  /* determine cache block state (RO, RW) */
  switch(block_RPC->nfsprogid) {
  case 2:
    switch(block_RPC->nfsprocid) {
    case NFSPROC_LOOKUP:
      block_RPC->state = ST_RO_CLEAN;
      break;
    case NFSPROC_GETATTR:
      block_RPC->state = ST_RO_CLEAN;
      break;
    default:
      block_RPC->state = ST_INV;
    }
    break;
  case 3:
  default:
    block_RPC->state = ST_INV;
  }

  /* write block into cache */
  written = write(fd, block_RPC, sizeof(*block_RPC));
  if(written != sizeof(*block_RPC)) {
    Dprintf(L_ERROR,"Could not write into cache!\n");
    flock(fd,LOCK_UN);
    close(fd);
    return -1;
  }
  /* release exclusive lock  */
  if ((flock(fd,LOCK_UN))==-1) {
    Dprintf(D_ACACHE,
	    "insert_in_cache: could not unlock cache bank %s for writing\n",
	    scratch_path);
    return -1;
  }
  close(fd);
  return 0;
}

/*************************************************************************
 * Initialize global variables                                           *
 * path    : base path to file cache directory                           *
 * return values:                                                        *
 * 0 : successfull                                                       *
 * -1 : unsuccessful                                                     *
 *************************************************************************/
int init_attr_cache_global(char *basepath, char *sessionid) {
  int staterr, mkdirerr;

  if (!(scratch_path = (char *)malloc(sizeof(*scratch_path)*MAX_STR)))
    return -1;

  if (!(SessionID = (char *)malloc(sizeof(*SessionID)*MAX_STR)))
    return -1;
  if(strlen(sessionid)>(sizeof(*SessionID)*MAX_STR))
    return -1;
  strcpy(SessionID,sessionid);

  if (!(BaseCachePath = (char *)malloc(sizeof(*BaseCachePath)*MAX_STR)))
    return -1;
  if(strlen(basepath)>(sizeof(*BaseCachePath)*MAX_STR))
    return -1;
  strcpy(BaseCachePath,basepath);

  if (!(statbuf = (struct stat *) malloc (sizeof(*statbuf))))
    return -1;

  if (!(block_scratch_RPC = (cacheblock *)malloc(sizeof(*block_scratch_RPC))))
    return -1;

  if (!(block_scratch_disk =(cacheblock *)malloc(sizeof(*block_scratch_disk))))
    return -1;

  /* check the existence of base path and session ID directories */
  staterr = stat(BaseCachePath, statbuf);
  if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
       (statbuf->st_mode & S_IWUSR))) {
    Dprintf(L_ERROR, "BaseCachePath directory %s does not exist\n",
	    BaseCachePath);
    return -1;
  }
  if((strlen(BaseCachePath)+1+strlen(SessionID))>MAX_STR) {
    Dprintf(L_ERROR,"cache block path too long\n");
    return -1;
  }
  sprintf(scratch_path,"%s/%s",BaseCachePath,SessionID);
  staterr = stat(scratch_path, statbuf);  
  if(!((staterr==0) && (statbuf->st_mode & S_IRUSR) &&
       (statbuf->st_mode & S_IWUSR))) {
    mkdirerr = mkdir(scratch_path,S_IRWXU | S_IRWXG);
    if(mkdirerr==-1) {
      Dprintf(L_ERROR, "Could not create SessionID directory %s\n",
	      scratch_path);
      return -1;
    }
  }
  return 0;
}

void dump_attr_cache_parameters(FILE *fp) {
  fprintf(fp,"Attribute cache parameters:\n");
  fprintf(fp,"\tCache size: %lu blocks\n",cache_size);
  fprintf(fp,"\tCache assoc: %du\n",cache_assoc);
  fprintf(fp,"\t# cache banks: %du\n",cache_banks);
  fprintf(fp,"\t# cache sets: %lu\n",cache_sets);
  fprintf(fp,"\t# sets per bank: %lu\n",sets_per_bank);
  fprintf(fp,"\tSize of cache block: %i\n",sizeof(*block_scratch_RPC));
  fprintf(fp,"\tSize of cache bank file: %lu\n",
	  sets_per_bank*cache_assoc*sizeof(*block_scratch_RPC));
}

/*************************************************************************
 * Initialize cache parameters                                           *
 * size    : cache size (# entries, pwr of 2)                            *
 * assoc   : cache associativity (pwr of 2)                              *
 * return values:                                                        *
 * 0 : successfull                                                       *
 *-1 : unsuccessful                                                      *
 *************************************************************************/
int init_attr_cache(unsigned long int size, int assoc, int banks,
		    char *basepath, char *sessionid) {

  unsigned long int localsize, localassoc, localbanks;

  if(basepath==NULL || sessionid==NULL) {
    Dprintf(L_ERROR, "basepath/sessionid error\n");
    return -1;
  }

  if(size==0 || size > MAX_CACHE_SIZE)
    return -1;
  if (assoc==0 || assoc > MAX_CACHE_ASSOC)
    return -1;

  if(init_attr_cache_global(basepath,sessionid) == -1)
    return -1;

  localsize=size;
  while (localsize>1) {
    if(localsize%2)
      return -1;
    localsize>>=1;
  }

  localassoc=assoc;
  while (localassoc>1) {
    if(localassoc%2)
      return -1;
    localassoc>>=1;
  }

  localbanks=banks;
  while (localbanks>1) {
    if(localbanks%2)
      return -1;
    localbanks>>=1;
  }

  cache_size = size;
  cache_assoc = assoc;
  cache_banks = banks;
  cache_sets = cache_size/cache_assoc;
  sets_per_bank = cache_size/(cache_assoc*cache_banks);
  Dprintf(D_DCACHE, "attr cache init successful\n"
                    "size: %d, assoc: %d, banks: %d "
		    "sets: %d, sets per bank: %d\n",
		    cache_size, cache_assoc, cache_banks,
		    cache_sets, sets_per_bank);
		      
  return 0;
}

/*************************************************************************
 * lookup block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : miss                                                              *
 * 1 : hit                                                               *
 * -1: error                                                             *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of diropargs                         *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *  res contains RPC result looked up from cache                         *
 *************************************************************************/
int cache_lookup(int prog_index, int proc_index, void *rpcargs, void *res) {
  int match=0;

  switch(prog_index) {
  case 2:
    switch(proc_index) {
    case NFSPROC_LOOKUP:
      /* create cache_scratch_RPC from rpcargs */
      if((argument_lookup_v2((diropargs *)rpcargs, block_scratch_RPC))==-1) {
	Dprintf(D_ACACHE,"cache_lookup: argument copy failed\n");
	return -1;
      }
      if((match=tag_match_and_inv(block_scratch_RPC,
				  block_scratch_disk, 0))==-1) {
	Dprintf(D_ACACHE,"cache_lookup: tag match failed\n");
	return -1;
      }
      if (match == 1) { /* hit; copy data */
	if((response_from_disk_cache_lookup_v2(block_scratch_disk,
					       (diropres *)res))==-1) {
	  Dprintf(D_ACACHE,"cache_lookup: response copy failed\n");
	  return -1;
	}
      }
      break;
    case NFSPROC_GETATTR:
      /* create cache_scratch_RPC from rpcargs */
      if((argument_getattr_v2((nfs_fh *)rpcargs, block_scratch_RPC))==-1) {
	Dprintf(D_ACACHE,"cache_lookup: argument copy failed\n");
	return -1;
      }
      if((match=tag_match_and_inv(block_scratch_RPC,
				  block_scratch_disk, 0))==-1) {
	Dprintf(D_ACACHE,"cache_lookup: tag match failed\n");
	return -1;
      }
      if (match == 1) { /* hit; copy data */
	if((response_from_disk_cache_getattr_v2(block_scratch_disk,
					       (attrstat *)res))==-1) {
	  Dprintf(D_ACACHE,"cache_lookup: response copy failed\n");
	  return -1;
	}
      }
      break;
    } /* switch (proc_index) */
    break;
  } /* switch(prog_index) */
  return match;
}

/*************************************************************************
 * insert block in cache                                                 *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : inserted                                                          *
 *-1 : did not insert                                                    *
 * side-effects:                                                         *
 *  block_scratch_RPC contains copy of results                           *
 *  block_scratch_disk contains copy of results in case of a hit         *
 *************************************************************************/
int cache_insert(int prog_index, int proc_index, void *res) {

  switch(prog_index) {
  case 2:
    switch(proc_index) {
    case NFSPROC_LOOKUP:
      /* create cache_scratch_RPC from rpcargs */
      if((response_remote_RPC_lookup_v2((diropres *)res,
					block_scratch_RPC))==-1) {
	Dprintf(D_ACACHE,"cache_insert: RPC response copy failed\n");
	return -1;
      }
      if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
	Dprintf(D_ACACHE,"cache_insert: insertion failed\n");
	return -1;
      }
      return 0;
    case NFSPROC_GETATTR:
      /* create cache_scratch_RPC from rpcargs */
      if((response_remote_RPC_getattr_v2((attrstat *)res,
					block_scratch_RPC))==-1) {
	Dprintf(D_ACACHE,"cache_insert: RPC response copy failed\n");
	return -1;
      }
      if((insert_in_cache(block_scratch_RPC, block_scratch_disk))==-1) {
	Dprintf(D_ACACHE,"cache_insert: insertion failed\n");
	return -1;
      }
      return 0;
    } /* switch(proc_index) */
    break;
  } /* switch(prog_index) */
  Dprintf(D_ACACHE,"cache_insert: unknown RPC call\n");
  return -1;
}

/*************************************************************************
 * Lookup entry for diroparg *arg in attribute cache.
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_diroparg_v2(diropargs *arg) {
  int match;

  if(arg==NULL) {
    Dprintf(D_ACACHE,"invalidate_diroparg_v2: arg==NULL\n");
    return -1;
  }

  /* copy diropargs to block_scratch_RPC */
  if((argument_lookup_v2(arg, block_scratch_RPC))==-1) {
    Dprintf(D_ACACHE,"invalidate_diroparg_v2: argument copy failed\n");
    return -1;
  }

  if((match=tag_match_and_inv(block_scratch_RPC,
			      block_scratch_disk, 1))==-1) {
    Dprintf(D_ACACHE,"invalidate_diroparg_v2: tag invalidate failed\n");
    return -1;
  }

  return match;
}

/*************************************************************************
 * Lookup entry for nfs_fh *arg in attribute cache.                      *
 * if found, invalidate it
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *************************************************************************/
int invalidate_nfs_fh_v2(nfs_fh *arg) {
  int match;

  if(arg==NULL) {
    Dprintf(D_ACACHE,"invalidate_nfs_fh_v2: arg==NULL\n");
    return -1;
  }

  /* copy diropargs to block_scratch_RPC */
  if((argument_getattr_v2(arg, block_scratch_RPC))==-1) {
    Dprintf(D_ACACHE,"invalidate_nfs_fh_v2: argument copy failed\n");
    return -1;
  }

  if((match=tag_match_and_inv(block_scratch_RPC,
			      block_scratch_disk, 1))==-1) {
    Dprintf(D_ACACHE,"invalidate_nfs_fh_v2: tag invalidate failed\n");
    return -1;
  }

  return match;
}

/*************************************************************************
 * check if need to invalidate block in cache                            *
 * prog_index, proc_index: type of RPC call                              *
 * argument = (void *)rpcargs from incoming RPC call                     *
 * return values:                                                        *
 * 0 : no invalidation needed                                            *
 * 1 : successful invalidation                                           *
 * -1: error                                                             *
 *                                                                       *
 * The following invalidations are performed:                            *
 *  NFS-V2:                                                              *
 *    CREATE, MKDIR: invalidate (diropargs *)rpcargs->where d            *
 *    REMOVE, RMDIR: invalidate (diropargs *)rpcargs                     *
 *    RENAME : invalidate (diropargs *)rpcargs->from, rpcargs->to        *
 *    SYMLINK: invalidate (diropargs *)rpcargs->from                     *
 *    LINK:    invalidate (diropargs *)rpcargs->to                       *
 *    SETATTR: invalidate (sattrstat *)rpcargs->file                     *
 *    WRITE:   invalidate (sattrstat *)rpcargs->file
 * side-effects:                                                         *
 * block_scratch_RPC contains copy of rpcargs                            *
 *************************************************************************/
int cache_inval(int prog_index, int proc_index, void *rpcargs) {
  createargs *crargs;
  renameargs *renargs;
  symlinkargs *symargs;
  linkargs *largs;
  sattrargs *sargs;
  writeargs *wargs;
  diropargs  *invalidate_arg1,*invalidate_arg2;
  int invres1=0, invres2=0;

  switch(prog_index) {
  case 2: /* NFS-V2 */
    /* first, set up invalidate_arg1 and invalidate_arg2 */
    /* most calls will only require one invalidation; in these
       cases, arg2 will be set to NULL. The exception is RENAME */
    switch(proc_index) {
    case NFSPROC_CREATE:
    case NFSPROC_MKDIR:
      crargs = (createargs *)rpcargs;
      invalidate_arg1 = (diropargs *)&crargs->where;
      invalidate_arg2 = NULL;
      invres1 = invalidate_nfs_fh_v2(&invalidate_arg1->dir);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      break;
    case NFSPROC_REMOVE:
    case NFSPROC_RMDIR:
      invalidate_arg1 = (diropargs *)rpcargs;
      invalidate_arg2 = NULL;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      break;
    case NFSPROC_RENAME:
      renargs = (renameargs *)rpcargs;
      invalidate_arg1 = (diropargs *)&renargs->from;
      invalidate_arg2 = (diropargs *)&renargs->to;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      if(memcmp(&invalidate_arg1->dir,&invalidate_arg2->dir,NFS_FHSIZE) != 0) {
	invres2 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg2->dir);
	if(invres2==-1) {
	  Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	  return -1;
	}
	Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres2);
      }
      break;
    case NFSPROC_SYMLINK:
      symargs = (symlinkargs *)rpcargs;
      invalidate_arg1 = (diropargs *)&symargs->from;
      invalidate_arg2 = NULL;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      break;
    case NFSPROC_LINK:
      largs = (linkargs *)rpcargs;
      invalidate_arg1 = (diropargs *)&largs->to;
      invalidate_arg2 = NULL;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&invalidate_arg1->dir);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      } 
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      break;
    case NFSPROC_SETATTR:
      /* SETATTR, WRITE are unlike the above calls; they invalidate files
	 based on the file handle rpcargs->file
	 They are handled separately here */
      sargs = (sattrargs *)rpcargs;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&sargs->file);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      return invres1;
      break;
    case NFSPROC_WRITE:
      wargs = (writeargs *)rpcargs;
      invres1 = invalidate_nfs_fh_v2((nfs_fh *)&wargs->file);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: invalidate fh %i\n", invres1);
      return invres1;
      break;
    default:
      /* no invalidations necessary for other calls */
      invalidate_arg1 = NULL;
      invalidate_arg2 = NULL;
    }
    /* now perform actual invalidation */
    if(invalidate_arg1!=NULL) {
      invres1 = invalidate_diroparg_v2(invalidate_arg1);
      if(invres1==-1) {
	Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	return -1;
      }
      Dprintf(D_ACACHE,"cache_inval: result of argument1: %i\n", invres1);
      if(invalidate_arg2!=NULL) {
	invres2 = invalidate_diroparg_v2(invalidate_arg2);
	if(invres2==-1) {
	  Dprintf(D_ACACHE,"Error invalidating attribute cache block\n");
	  return -1;
	}
	Dprintf(D_ACACHE,"cache_inval: result of argument2: %i\n", invres2);
      }
      return((invres1+invres2)>0);
    }
    
    break; /* case 2 */
  case 3: /* NFS-V3 */
    break;
  default:
    return -1; /* unknown NFS version */
  }
  return -1;
}
