/*
 * Copyright (c) 2002-2005, In-VIGO Team
 * Advanced Computing and Information Systems (ACIS) Laboratory
 * University of Florida
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * mountd_clnt.c	This file has funtions that handle the
 *                      creation of the client-side of the gvfs proxy
 *
 * Author: Renato J. Figueiredo
 */
 
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <rpc/rpc.h>
#include <rpc/key_prot.h>
#include <rpc/pmap_clnt.h>
#ifdef SYSV
#include <rpc/clnt_soc.h>
#endif
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "nfs_prot.h"
#include "mountd.h"

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/clnt_soc.h>
#endif /* SOLARIS */

struct timeval timeout = { 180, 0 }; /* default time out */
struct timeval retry_timeout = { 20, 0 }; /* retry interval = 10s */

/*
 * Setup a connection to host "svr", program "prog" and version "vers"
 * using a privileged port.
 */
int
setup(int type, struct sockaddr_in *svr, int prog, int vers, int rootsock)
{
  int s;

  if(rootsock==0) {
    s = RPC_ANYSOCK;
  } else { /* privileged port (rootsock) has been inherited */
    s = rootsock;
  }
  return s;
}

/*
 * Returns an auth handle with parameters determined by doing lots of
 * syscalls.
 */
AUTH *
create_authenticator(int authtype)
{
  if (authtype == AUTH_UNIX) {
    return authunix_create_default();
  } else {
    fprintf(stderr, "create_authenticator: no secure nfs support\n");
    exit(-1);
  }
}

/****************************************************************************
 * Connect the client to a mount server
 * mounthost : the server's address
 * mountport : the port it listens to
 * vers    : the version of the NFS protocol to connect to
 * rootsock: ordinarily, zero; non-zero if there is a pre-opened
 *           privileged socket in which the connection should be
 *           established (rootsock is the ID of the preopened
 *           socket in this case)
 ****************************************************************************/
CLIENT *
open_mountdclient(char *mounthost, int mountport, int vers, int rootsock,
		  int flags)
{
  int proto, sock;
  struct sockaddr_in mountserver_addr; /* remote mount server address */
  CLIENT *retclnt = NULL;

  /* handle numeric and string IP addresses */
  if (isdigit((int)*mounthost)) {
    mountserver_addr.sin_addr.s_addr = inet_addr(mounthost);
  } else {
    struct hostent *hp = gethostbyname(mounthost);
    if (hp == NULL) {
      fprintf(stderr, "%s: unknown host\n", mounthost);
      return NULL;
    }
    memcpy(&mountserver_addr.sin_addr.s_addr, hp->h_addr, hp->h_length);
    mounthost = (char *) hp->h_name;
  }
  mountserver_addr.sin_family = AF_INET;
  mountserver_addr.sin_port = ntohs(mountport);

  /* Set up a UDP/TCP proxy mount client and connected to the server
     If successful, retclnt is a pointer to the RPC mount client */
  switch (flags & TRANSPORT_MASK) {
  case NFS_OVER_UDP:
    Dprintf(D_CALL,"mountd_client: opening UDP connection\n");
    /* try using NFS over UDP (standard Sun setup) */
    proto = IPPROTO_UDP;
    sock = setup(SOCK_DGRAM, &mountserver_addr, MOUNT_PROGRAM,
		 vers, rootsock);
    if ((retclnt = clntudp_bufcreate(&mountserver_addr,
				     MOUNT_PROGRAM, vers, timeout,
				     &sock, NFS_MAXDATA,
				     NFS_MAXDATA)) == (CLIENT *)0) {
      clnt_pcreateerror("mount clntudp_create");
      if (sock != RPC_ANYSOCK)
	close(sock);
      return NULL;
    }
    break;
  case NFS_OVER_TCP:
    Dprintf(D_CALL,"mountd_client: opening TCP connection\n");
    /* try using NFS over TCP (standard Sun setup) */
    proto = IPPROTO_TCP;
    sock = setup(SOCK_STREAM, &mountserver_addr, MOUNT_PROGRAM,
		 vers, rootsock);
    if ((retclnt = clnttcp_create(&mountserver_addr,
				  MOUNT_PROGRAM, vers, &sock,
				  NFS_MAXDATA, NFS_MAXDATA))
	== (CLIENT *)0) {
      clnt_pcreateerror("mount clnttcp_create");
      if (sock != RPC_ANYSOCK)
	close(sock);
      return NULL;
    }
    break;
  default:
    Dprintf(L_FATAL,"mountd_client: unknwon transport\n");
    break;
  }

  /* set timeouts */
  clnt_control(retclnt, CLSET_TIMEOUT, (char *)&timeout);
  clnt_control(retclnt, CLSET_RETRY_TIMEOUT, (char *)&retry_timeout);

  /* set authentication credentials */
  retclnt->cl_auth = create_authenticator(AUTH_UNIX);
  return retclnt;
}

/****************************************************************************
 * Close proxy NFS-mount client
 ****************************************************************************/
void
close_mountdclient(CLIENT *retclnt)
{
    if (retclnt) {
        auth_destroy(retclnt->cl_auth);
        clnt_destroy(retclnt);
    }
}


/****************************************************************************
 * Returns an auth handle for the client, with the option of
 * overriding uid/gid credentials.
 ***************************************************************************/

AUTH *create_authenticator_with_map(int authtype, int uid, int gid)
{
  gid_t gids[1];
  char machname[1024];

  if (authtype == AUTH_UNIX) {
    gids[0] = gid;
    if((gethostname(machname,1024))==-1) {
      Dprintf(L_ERROR, "create authenticator: cannot obtain host name\n");
      return(NULL);
    }
    return authunix_create(machname, uid, gid, 1, gids);
  } else {
    Dprintf(L_ERROR, "create_authenticator: no secure nfs support\n");
    return(NULL);
  }
}
