/*
 * nfsd		This is the main program for the grid virtual file
 *              system proxy gvfs.nfsd. This proxy receives, modifies
 *              and forwards NFS RPC calls. 
 *
 * Author: 	Renato J. Figueiredo
 *         	Ming Zhao 
 *
 * Updated: 	Jan., 2004
 *		Added cache support situations that requested data in READ
 *		cross block boudaries;
 *		Corrected the invalidation of data cache when requested data
 *		crosses block boundaries;
 *		Corrected the problem that seekset is not evaluated properly
 *		in dfind_replace_frame function.
 *              Added support for session key authentication.
 *
 * Updated:     Feb., 2005
 *              Corrected the problem that id mapping is not done within the
 *              arguments of setattr call.
 *
 * Based on User-level NFSD code by:
 *
 * Authors:	Mark A. Shand, May 1988
 *		Donald J. Becker, <becker@super.org>
 *		Rick Sladkey, <jrs@world.std.com>
 *		Fred N. van Kempen, <waltje@uWalt.NL.Mugnet.ORG>
 *		Eric Kasten, <tigger@tigger.cl.msu.edu>
 *		Olaf Kirch, <okir@monad.swb.de>
 *
 *		Copyright 1988 Mark A. Shand
 *		This software maybe be used for any purpose provided
 *		the above copyright notice is retained.  It is supplied
 *		as is, with no warranty expressed or implied.
 */

#include "nfsd.h"
#include <rpc/pmap_clnt.h>
#include "getopt.h"
#include "rpcmisc.h"
#include "signals.h"
#include "cache_attrib.h"
#include "cache_data.h"
#include "sk.h"
#ifdef __linux__ /* XXX - MvS: for UNIX sockets. */
#  include <sys/un.h>
#endif
#ifdef HAVE_SYSLOG_H
# include <syslog.h>
#endif

#ifdef SOLARIS
#define PORTMAP
#include <rpc/rpc.h>
#include <rpc/svc_soc.h>
#endif /* SOLARIS */

#define MULTIPLE_SERVERS

/* Flags for auth_fh */
#define CHK_READ	0
#define CHK_WRITE	1
#define CHK_NOACCESS	2

extern char version[];
static char *program_name;

/*
 * Option table for command-line arguments
 */
static struct option longopts[] = {
      { "debug",		required_argument,	0,	'd' },
      { "foreground",		0,			0,	'F' },
      { "exports-file",		required_argument,	0,	'f' },
      { "help",			0,			0,	'h' },
      { "host-server",          required_argument,      0,      'H' },
      { "portin",      		required_argument,	0,	'P' },
      { "portout",      	required_argument,	0,	'O' },
      { "sock",      	        required_argument,	0,	'S' },
      { "transport",      	required_argument,	0,	't' },
      { "nfsversion",      	required_argument,	0,	'V' },
      { "cookiefile",	        required_argument,	0,	'C' },
      { "cachesessionid",	required_argument,	0,	'i' },
      { "enableattrcache",      0,	                0,	'A' },
      { "enabledatacache",      0,	                0,	'D' },
      { "cachebasepath",       	required_argument,	0,	'c' },
      { "attrcachesize",       	required_argument,	0,	'z' },
      { "attrcacheassoc",	required_argument,	0,	'a' },
      { "attrcachebanks",	required_argument,	0,	'b' },
      { "datacachesize",       	required_argument,      0,	'e' },
      { "datacacheassoc",	required_argument,	0,	'o' },
      { "datacachebanks",	required_argument,	0,	'k' },
      { "syncinterval", 	0,                      0,      'T' },
      { "sessionkeysend",	0,			0,	's' },
      { "sessionkeymatch",	0,			0,	'm' },
      { "promiscuous",		0,			0,	'p' },
      { "version",		0,			0,	'v' },
      { NULL,		0,	0, 0 }
};
static const char * shortopts = "d:Ff:hH:P:O:S:t:V:C:i:ADc:z:a:b:e:o:k:T:s:m:pv::";

/* Pointer to NFS client authenticated by the proxy; obtained from
   auth_clnt() in auth_clnt.c */
nfs_client *		nfsclient = NULL;
/* Pointer to mount point exported by the proxy; initialized from
   exports file by auth_init() in auth_init.c */
nfs_mount *		nfsmount = NULL;
/* Global variable used for SIGHUP handling; SIGHUPs force the proxy
   to re-read its configuration files; this variable implements a
   simple form of mutual exclusion and makes sure this
   does not happen while an RPC request is being serviced */	
int			need_reinit = 0;	
/* Global variable used for file cache flushes; this variable
   implements a simple form of mutual exclusion and makes sure this
   does not happen while an RPC request is being serviced */
int			need_flush = 0;
/* Global variable that determines if attribute caching is enabled */
int                     attr_cache_enabled = 0;
/* Global variable that determines if data caching is enabled */
int                     data_cache_enabled = 0;
/* Global variables that determines if auth_clnt.c must refresh client
   UNIX AUTH authenticator */
int                     refresh_authentication_handle_uid = -1;
int                     refresh_authentication_handle_gid = -1;
/* Global variable: current access hit the data or attr cache */
int                     data_attr_cache_hit = 0;
/* Buffer for cached data replies */
char                    data_cache_buffer[MAX_NFS_BLOCKSIZE];
/* NFS version that this proxy will forward */
int                     nfsvers = 2;
int                     authvers;
/* Global variable that determines if periodical sync is enabled */
int			sync_enabled = 0;;
int                     sync_interval; /* defines client polling interval */
int			need_alarm = 0;

/* Call profiling defines for functions and data structures; see
   Makefile.in for a description of the profiling #define's */

#ifdef CALL_PROFILING
extern void	dump_stats(int sig);
extern struct timeval inittime;
#endif

CLIENT *authclient1;

/* NFSD client for forwarding V2 NFS protocol calls */
CLIENT *nfsdclient2;           
CLIENT *authclient2;

/* NFSD client for forwarding V3 NFS protocol calls */
#ifdef NFSV3
CLIENT *nfsdclient3;
CLIENT *authclient3;
#endif /* NFSV3 */

/* Function definitions */
nfs_mount *auth_fh(struct svc_req *rqstp, nfsstat *statp);
static void usage(FILE *fp, int n);
static void terminate(void);
static RETSIGTYPE sigterm(int sig);
static RETSIGTYPE sigusr2(int sig);
RETSIGTYPE flush_caches(int sig);
RETSIGTYPE reinitialize(int sig);
RETSIGTYPE setalarm(int sig);

/* External function definitions */
extern CLIENT *open_nfsdclient(char *nfshost, int nfsport, 
			       int vers, int rootsock, int flags);
extern void close_nfsdclient(CLIENT *retclnt);
extern int flush_attr_cache(void);
extern int flush_data_cache(void);
extern void log_cache_hits(int proc_index);
extern int invalidate_diroparg_v2(diropargs *arg);

/* macros for handling NFS requests */
#define AUTHENTICATE_CLIENT \
        if((nfsmount=auth_fh(rqstp, &astat))==NULL) { \
          return(astat);}
#define FORWARD_REQUEST(CLIENT_FUNCTION,CLIENT,MESSAGE) \
        if ((res = CLIENT_FUNCTION(argp,CLIENT)) == NULL) { \
           clnt_perror(CLIENT,MESSAGE); \
	   return NFSERR_ACCES; \
        } 
#ifdef NFSV3
#define FORWARD_REQUEST3(CLIENT_FUNCTION,CLIENT,MESSAGE) \
        if ((res = CLIENT_FUNCTION(argp,CLIENT)) == NULL) { \
           clnt_perror(CLIENT,MESSAGE); \
	   return NFS3ERR_ACCES; \
        };
#endif /* NFSV3 */


/****************************************************************************
 * Map user and group IDs of attributes returned by the NFS server
 *  back to the user/group IDs of the client.
 *  The original IDs (coming from a client request) are stored in
 *  *attr; the mapping functions are defined in ugid_map.c; the
 *  mapped IDs (to be forwarded by the proxy) overwrite the
 *  client IDs
 ****************************************************************************/
void map_return_attributes(fattr *attr, nfs_mount *mp, 
		     struct svc_req *rqstp)
{
  int serveruid, servergid, clientuid, clientgid;

  serveruid = attr->uid;
  servergid = attr->gid;
  clientuid = ruid(serveruid, nfsmount, rqstp);
  clientgid = rgid(servergid, nfsmount, rqstp);
  attr->uid = clientuid;
  attr->gid = clientgid;
  Dprintf(D_UGID, "map_return_attributes: (%i,%i) -> (%i,%i)\n",
	  serveruid,servergid,clientuid,clientgid);
}

#ifdef NFSV3
void map_return_attributes3(fattr3 *attr, nfs_mount *mp,
		       struct svc_req *rqstp)
{
  uint32 serveruid, servergid, clientuid, clientgid;

  serveruid = attr->uid;
  servergid = attr->gid;
  clientuid = (uint32) ruid(serveruid, nfsmount, rqstp);
  clientgid = (uint32) rgid(servergid, nfsmount, rqstp);
  attr->uid = clientuid;
  attr->gid = clientgid;
  Dprintf(D_UGID, "map_return_attributes3: (%i,%i) -> (%i,%i)\n",
	  serveruid,servergid,attr->uid,attr->gid);
}
#endif /* NFSV3 */

/****************************************************************************
 * Return NFS error string
 ****************************************************************************/
char *
nfs_error(enum nfsstat stat)
{
  switch (stat) {
  case NFS_OK:
    return "No error";
  case NFSERR_PERM:
    return "Not owner";
  case NFSERR_NOENT:
    return "No such file or directory";
  case NFSERR_IO:
    return "I/O error";
  case NFSERR_NXIO:
    return "No such device or address";
  case NFSERR_ACCES:
    return "Permission denied";
  case NFSERR_EXIST:
    return "File exists";
  case NFSERR_NODEV:
    return "No such device";
  case NFSERR_NOTDIR:
    return "Not a directory";
  case NFSERR_ISDIR:
    return "Is a directory";
  case NFSERR_FBIG:
    return "File too large";
  case NFSERR_NOSPC:
    return "No space left on device";
  case NFSERR_ROFS:
    return "Read-only file system";
  case NFSERR_NAMETOOLONG:
    return "File name too long";
  case NFSERR_NOTEMPTY:
    return "Directory not empty";
  case NFSERR_DQUOT:
    return "Disc quota exceeded";
  case NFSERR_STALE:
    return "Stale NFS file handle";
  case NFSERR_WFLUSH:
    return "Write cache flushed";
  default:
    return "UKNOWN NFS ERROR";
  }
}

/****************************************************************************
 * auth_fh
 *
 * This function authenticates the file handle provided by the client.
 * If the file handle is authenticated by this function, it returns
 * a '1", otherwise a "0". Requests that are authenticated are
 * forwarded by the proxy; requests not authenticated are not
 * forwarded.
 ****************************************************************************/
nfs_mount * 
auth_fh(struct svc_req *rqstp, nfsstat *statp)
{
  nfs_mount *scan = NULL;

  /* First, authenticate the client. If not successfull, flag NFS
     access error and return zero */
  if ((nfsclient = auth_clnt(rqstp)) == NULL) {
    *statp = NFSERR_ACCES;
    return 0;
  }

  /* Authenticate by the session key if it is enabled. Successful 
     authentication returns client id larger or equal than 1 */
  if(match_sk_enabled)
    if((client_id = auth_sk(rqstp)) < 1) {
      *statp = NFSERR_ACCES;
      return 0;      
  }

  /* Make sure mount point is non-null */
  if (nfsclient->m == NULL) {
    Dprintf(L_ERROR, "NULL mountpoint\n");
    *statp = NFSERR_ACCES;
    return 0;
  }

  /* Authenticate user (uid/gid); return NFS access error if user
     is not authorized */
  *statp = NFSERR_ACCES;
  for(scan = nfsclient->m; scan != NULL; scan = scan->next) {
    if(auth_user(scan, rqstp)) {
      /* user is authorized to access this mount point 
      mp = scan; 
      fprintf(stderr,"scaning: nfsmount = 0x%x, scan = 0x%x\n",nfsmount,scan);
      fflush(stderr);
      */
      *statp = NFS_OK;
      break;
    }
  }

  if(scan == NULL)
    return scan;

  /* Check if request comes from insecure port and not explicitly
     allowed for this mount point (i.e. option "insecure" not set
     in exports file) */
  if (!allow_non_root && scan->o.secure_port
      && !SECURE_PORT(svc_getcaller(rqstp->rq_xprt)->sin_port)) {
    Dprintf(L_ERROR,
	    "NFS request from %s originated on insecure port, %s\n",
	    nfsclient->clnt_name,
	    "psychoanalysis suggested");
    *statp = NFSERR_ACCES;
    return 0;
  }

  /* Depending on value of *statp, auth has been successful or not */
  /* return mount pointer (NULL if not successfull...)*/
  return(scan);
}

/****************************************************************************
 * The functions below are called when the server-side of the proxy
 * receives an NFS request. There are as many functions as defined
 * in the NFS protocol (for NFS-V2, and NFS-V3 if enabled). All functions
 * have a similar structure, as follows:
 *   - 1. authenticate request with auth_fh; return immediately
 *     if not authenticated.
 *   - 2. forward the request (via a client call from the proxy).
 *     Note that requests are forwarded with the uid/gid of the
 *     user that is executing the proxy.
 *   - 3. check for errors in the processing of the forwarded request;
 *     map the server return attributes, and return the mapped NFS
 *     reply to the client
 * The "wrappers" of the following functions came from `rpcgen -l nfs_prot.x`.
 * This normally generates the client routines, but it provides nice
 * prototypes for the server routines also.
 * Refer to the NFS protocol for descriptions of the procedures and
 * data structures used in these functions
 ****************************************************************************/

int
nfsd_nfsproc_null_2(void *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_null_2, nfsdclient2, "nfsproc_null_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* No need to map attributes in NULL call */
  return (NFS_OK);
}

int
nfsd_nfsproc_getattr_2(nfs_fh *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;
  int lkupres=0, insres;

  Dprintf(D_DCACHE, "*** GETATTR:\n");
  hexdump(argp->data, NFS_FHSIZE, D_DCACHE);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
    lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc, 
			   (void *)argp, (void *)&result.attrstat);
    if(lkupres==1) {
      data_attr_cache_hit=1;
#ifdef CALL_PROFILING /* log cache hit */
      log_cache_hits(rqstp->rq_proc);
#endif
    }
    Dprintf(D_ACACHE, "cache lookup: getattr_2(%i,%i) = %s\n",
	    rqstp->rq_vers, rqstp->rq_proc,
	    (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));
  }
  if(lkupres!=1) { /* cache disabled or cache miss; forward request to server */
    /* Forward request to server */
    FORWARD_REQUEST(nfsproc_getattr_2, nfsdclient2, "nfsproc_getattr_2");
    if (res->status!=NFS_OK) { 
      return res->status;
    }
    if(attr_cache_enabled) { /* insert RPC response in cache */
      insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)res);
      Dprintf(D_ACACHE, "cache insert: getattr_2(%i,%i) = %s\n",
	      rqstp->rq_vers, rqstp->rq_proc,
	      (insres==0 ? "OK":"ERROR"));
    }
    result.attrstat = *res;
  }

  /* Map attributes and return to client */
  map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
  return (result.attrstat.status);
}

int
nfsd_nfsproc_setattr_2(sattrargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;
  uid_t cuid;
  gid_t cgid;
  int insres;
  
  Dprintf(D_DCACHE, "*** SETATTR:\n");
  hexdump(argp->file.data, NFS_FHSIZE, D_DCACHE);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  if(attr_cache_enabled) /* invalidate attributes cache */
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);

  /* Forward request to server */
  if(argp->attributes.uid != -1) { 
    cuid = luid(argp->attributes.uid, nfsmount, rqstp);
    Dprintf(D_ACACHE, "map uid in args from %d to %d\n", argp->attributes.uid, cuid);    
    argp->attributes.uid = cuid;    
  }
  if(argp->attributes.gid != -1) { 
    cgid = lgid(argp->attributes.gid, nfsmount, rqstp);
    Dprintf(D_ACACHE, "map gid in args from %d to %d\n", argp->attributes.gid, cgid);
    argp->attributes.gid = cgid;   
  }
  FORWARD_REQUEST(nfsproc_setattr_2, nfsdclient2, "nfsproc_setattr_2");
  if(res->status != NFS_OK) {
    return res->status;
  }

  if(attr_cache_enabled) {
    insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)res);
    Dprintf(D_ACACHE, "cache insert: setattr_2(%i,%i) = %s\n",
	    rqstp->rq_vers, NFSPROC_GETATTR, (insres==0 ? "OK":"ERROR"));
  }			    

  /* Map attributes and return to client */
  result.attrstat = *res;
  map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
  return (result.attrstat.status);
}

int
nfsd_nfsproc_root_2(void *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_root_2, nfsdclient2, "nfsproc_root_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* Map attributes and return to client */
  result.attrstat = *res;
  map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
  return (result.attrstat.status); 
}

int
nfsd_nfsproc_lookup_2(diropargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  diropres *res;
  int lkupres=0, lkupres2=0, insres, invres;
  attrstat res_attr;

  Dprintf(D_DCACHE, "*** LOOKUP: %s\n", argp->name);
  hexdump(argp->dir.data, NFS_FHSIZE, D_DCACHE);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(attr_cache_enabled) { /* look up cache; retrieve results if hit */
    lkupres = cache_lookup(rqstp->rq_vers, rqstp->rq_proc, 
			   (void *)argp, (void *)&result.diropres);
    Dprintf(D_ACACHE, "cache lookup: lookup_2(%i,%i) = %s\n",
	    rqstp->rq_vers, rqstp->rq_proc,
	    (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));
	    
    if(lkupres) {  
      if (result.diropres.status != NFS_OK) {
        Dprintf(D_ACACHE, "file does not exist\n");
        Dprintf(D_ACACHE, "status: %d\n", result.diropres.status);	
#ifdef CALL_PROFILING
	log_cache_hits(rqstp->rq_proc);
#endif
	return result.diropres.status;      
      }

      lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR, 
  		   (void *)&result.diropres.diropres_u.diropres.file,
  		   (void *)&res_attr);
      Dprintf(D_ACACHE, "cache lookup: lookup_2(%i,%i) = %s\n",
	  rqstp->rq_vers, NFSPROC_GETATTR,
	  (lkupres2==1 ? "HIT" : (lkupres2==0 ? "MISS":"ERROR")));
	  
      /* keep consistent with GETATTR cache */
      if(lkupres2) { 
	data_attr_cache_hit=1;
#ifdef CALL_PROFILING
	log_cache_hits(rqstp->rq_proc);
#endif
        result.diropres.diropres_u.diropres.attributes = 
			res_attr.attrstat_u.attributes;
        goto out;
      }
      else { /* if not cached in GETATTR cache, then consider LOOKUP cache stale */
	invres = invalidate_diroparg_v2(argp);
	Dprintf(D_ACACHE, "proc_lookup_2: invalidate LOOKUP cache: %s\n",
			(invres==-1 ? "ERROR":"OK"));
      }	
    }
  }
  
  /* cache not enabled or cache miss */
  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_lookup_2, nfsdclient2, "nfsproc_lookup_2");
  if(res->status != NFS_OK) {
    if(attr_cache_enabled && res->status == NFSERR_NOENT) { 
      Dprintf(D_ACACHE, "file does not exist\n");
      cache_lookup(rqstp->rq_vers, rqstp->rq_proc, (void *)argp, (void *)&result.diropres);
      result.diropres = *res;
      insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)&result.diropres);
      Dprintf(D_ACACHE, "cache insert: lookup_2(%i,%i) = %s\n",
	      rqstp->rq_vers, rqstp->rq_proc,
	      (insres==0 ? "OK":"ERROR"));
    }
    return res->status;
  }
  result.diropres = *res;
  
  if(attr_cache_enabled) {
    insres = cache_insert(rqstp->rq_vers, rqstp->rq_proc, (void *)&result.diropres);
    Dprintf(D_ACACHE, "cache insert: lookup_2(%i,%i) = %s\n",
    		rqstp->rq_vers, rqstp->rq_proc,
    		(insres==0 ? "OK":"ERROR"));      

    lkupres2 = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR, 
    		(void *)&result.diropres.diropres_u.diropres.file,
    		(void *)&res_attr);
    if(!lkupres2) {
      res_attr.status = NFS_OK;
      res_attr.attrstat_u.attributes = 
      		result.diropres.diropres_u.diropres.attributes;
      insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&res_attr);
      Dprintf(D_DCACHE, "cache insert: lookup_2(%i,%i) = %s\n",
    		  rqstp->rq_vers, NFSPROC_GETATTR, (insres==0 ? "OK":"ERROR"));
    }
  }
    
out:
  hexdump(result.diropres.diropres_u.diropres.file.data, NFS_FHSIZE, D_DCACHE);
  /* Map attributes and return to client */
  map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
  return (result.diropres.status);
}

int
nfsd_nfsproc_readlink_2(nfs_fh *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  readlinkres *res;

  Dprintf(D_DCACHE, "*** READLINK:\n");
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_readlink_2, nfsdclient2, "nfsproc_readlink_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* No need to map attributes */
  result.readlinkres = *res;
  return (result.readlinkres.status);
}

int
nfsd_nfsproc_read_2(readargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  readres *res=NULL, read_res;
  int lkupres=0,insres, crossed_boundary = 0, match1 = 0, match2 = 0;
  int requestcnt, requestoffset;
  readargs args;
  attrstat res_attr, *res_attr_pt;
  
  Dprintf(D_DCACHE, "*** READ: offset %d count %d \n", argp->offset, argp->count);
  hexdump(argp->file.data, NFS_FHSIZE, D_DCACHE);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(!data_cache_enabled) goto no_caching;
  else { /* look up cache; retrieve results if hit */
    /* get attributes for validation of cached data */
    lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR, 
		   (void *)&argp->file, (void *)&res_attr);
    Dprintf(D_DCACHE, "cache lookup: read_2(%i,%i) = %s\n",
	  rqstp->rq_vers, NFSPROC_GETATTR,
	  (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));
    if(lkupres != 1) {   
      if ((res_attr_pt = nfsproc_getattr_2(&argp->file,nfsdclient2)) == NULL) { 
        Dprintf(D_DCACHE, "getattr from server error %d\n", res_attr_pt->status);
	goto no_caching;
      }
      if(res_attr_pt->status != NFS_OK) {
        clnt_freeres(nfsdclient2, (xdrproc_t)xdr_attrstat, (caddr_t)res_attr_pt);
	return res_attr_pt->status;
      }       
       /* insert attributes into cache */
      insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)res_attr_pt);
      Dprintf(D_DCACHE, "cache insert: read_2(%i,%i) = %s\n",
	      rqstp->rq_vers, NFSPROC_GETATTR, (insres==0 ? "OK":"ERROR"));
      res_attr = *res_attr_pt;
      clnt_freeres(nfsdclient2, (xdrproc_t)xdr_attrstat, (caddr_t)res_attr_pt);
      if(insres != 0) goto no_caching;
    }
    
    /* set up static buffer for cached data */
    result.readres.readres_u.reply.data.data_val = &data_cache_buffer[0];
    args = *argp;
    if(((args.offset % MAX_NFS_BLOCKSIZE) + args.count) > 
      	      MAX_NFS_BLOCKSIZE) { 
      /* read crossed boundary: lookup the first block */
      Dprintf(D_DCACHE, "read crossed boundary\n");
      crossed_boundary = 1;
      requestcnt = MAX_NFS_BLOCKSIZE - (args.offset % MAX_NFS_BLOCKSIZE);
      args.count = requestcnt;
      args.totalcount = requestcnt;
    }
    /* try to fetch locally */
    lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc,
			    (void *)&args, (void *)&result.readres);
    Dprintf(D_DCACHE, "cache lookup: read_2(%i,%i) = %s\n",
	    rqstp->rq_vers, rqstp->rq_proc,
	    (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));

    if(lkupres) {
      if(result.readres.readres_u.reply.attributes.mtime.seconds <
      		res_attr.attrstat_u.attributes.mtime.seconds) {
	Dprintf(D_DCACHE, "cached data may be stale; flush\n");
        dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
	lkupres = 0;
      }
    }
    
    if(lkupres == -1) goto no_caching; /* error */
    else if(lkupres == 0) { /* not cached */
      /* fetch the entire first block */
      args.offset = (argp->offset) - (argp->offset%MAX_NFS_BLOCKSIZE);
      args.count = MAX_NFS_BLOCKSIZE;
      args.totalcount = MAX_NFS_BLOCKSIZE;

      if(((res = nfsproc_read_2(&args,nfsdclient2)) == NULL) ||
         	      (res->status != NFS_OK)) {
  	 Dprintf(L_ERROR, "read from server error %d\n", res->status);
  	 goto no_caching;
      }  
      Dprintf(D_DCACHE, "forward: offset %d, return: data len %d\n",
      		args.offset, res->readres_u.reply.data.data_len);
      /* insert the first block into cache */
      if(res->readres_u.reply.data.data_len > 0) {
	args.offset = argp->offset;
	if(crossed_boundary)
	  args.count = MAX_NFS_BLOCKSIZE - (args.offset % MAX_NFS_BLOCKSIZE);
	else  args.count = argp->count;
	args.totalcount = args.count;
	insres = dcache_insert((void *)&args,rqstp->rq_vers,
      		  rqstp->rq_proc,(void *)res);
	Dprintf(D_DCACHE, "dcache insert: read_2(%i,%i) = %s\n",
	       rqstp->rq_vers, rqstp->rq_proc, (insres==0 ? "OK":"ERROR"));
	if(insres != 0) {
	  clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)res);
	  goto no_caching;
	}
      }
      result.readres = *res;
    }
    else match1 = 1; /* hit on the first block */
    
    if(crossed_boundary) { /* fetch second block */
      read_res = result.readres;
      if((read_res.readres_u.reply.data.data_val = malloc(argp->count))
      		== NULL) {
        Dprintf(D_DCACHE,  "error malloc memory\n");
	if(!match1)
	  clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)res);
	goto no_caching;
      }
      if(result.readres.readres_u.reply.data.data_len > 0)
	if(memcpy(read_res.readres_u.reply.data.data_val, 
      		  result.readres.readres_u.reply.data.data_val,
		  result.readres.readres_u.reply.data.data_len) == NULL) {
	  Dprintf(D_DCACHE, "error memory copy\n");
	  free(read_res.readres_u.reply.data.data_val);
	  if(!match1)
	    clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)&result.readres);
          goto no_caching;
	}
      read_res.readres_u.reply.data.data_len = 
      		  result.readres.readres_u.reply.data.data_len;
      if(!match1)
        clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)&result.readres);
      Dprintf(D_DCACHE, "got first part: data_len %d\n",
      		read_res.readres_u.reply.data.data_len);
		
      result.readres.readres_u.reply.data.data_val = &data_cache_buffer[0];
      requestoffset = argp->offset + argp->count;
      args.offset = requestoffset - (requestoffset % MAX_NFS_BLOCKSIZE);          
      args.count = requestoffset % MAX_NFS_BLOCKSIZE;
      args.totalcount = args.count;
      /* try to fetch locally */  
      lkupres = dcache_lookup(rqstp->rq_vers, rqstp->rq_proc, 
			      (void *)&args, (void *)&result.readres);
      Dprintf(D_DCACHE, "dcache lookup: read_2(%i,%i) = %s\n",
	      rqstp->rq_vers, rqstp->rq_proc,
	      (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));   
      if(lkupres) {
        if(result.readres.readres_u.reply.attributes.mtime.seconds <
      		  res_attr.attrstat_u.attributes.mtime.seconds) {
	  Dprintf(D_DCACHE, "the file has been modified on server, \
		  	     and cached data may be stale\n");
          dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)&args);
	  lkupres = 0;
	}
      }
      if(lkupres == -1) {  /* error */
        free(read_res.readres_u.reply.data.data_val);
	goto no_caching;
      }
      else if(lkupres == 0) { /* not cached */
	/* fetch the entire second block */
	args.count = MAX_NFS_BLOCKSIZE;
	args.totalcount = MAX_NFS_BLOCKSIZE;
	
	if(((res = nfsproc_read_2(&args,nfsdclient2)) == NULL) ||
         		(res->status != NFS_OK)) {
  	   Dprintf(L_ERROR, "read from server error %d\n", res->status);
	   free(read_res.readres_u.reply.data.data_val);
  	   goto no_caching;
        }
	Dprintf(D_DCACHE, "forward: offset %d, return: data len %d\n",
      		  args.offset, res->readres_u.reply.data.data_len);
	/* insert the second block into cache */	  
	if(res->readres_u.reply.data.data_len > 0) {
	  args.count = requestoffset % MAX_NFS_BLOCKSIZE;
	  args.totalcount = args.count;
	  insres = dcache_insert((void *)&args,rqstp->rq_vers,
      		    rqstp->rq_proc,(void *)res);
	  Dprintf(D_DCACHE, "dcache insert: read_2(%i,%i) = %s\n",
		 rqstp->rq_vers, rqstp->rq_proc, (insres==0 ? "OK":"ERROR"));
	  if(insres != 0) {
	    free(read_res.readres_u.reply.data.data_val);
	    clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)res);
	    goto no_caching;
	  }
	}
	result.readres = *res;
      }
      else match2 = 1; /* hit on the second block */
      
      if(result.readres.readres_u.reply.data.data_len > 0)
        if(memcpy(&read_res.readres_u.reply.data.data_val[read_res.readres_u.reply.data.data_len],
      		  result.readres.readres_u.reply.data.data_val,
		  result.readres.readres_u.reply.data.data_len) == NULL) {
	  Dprintf(D_DCACHE, "error memory copy\n");
	  free(read_res.readres_u.reply.data.data_val);
	  if(!match2)
	    clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)&result.readres);
          goto no_caching;
	}
      read_res.readres_u.reply.data.data_len += 
      		result.readres.readres_u.reply.data.data_len;     
      if(!match2)
        clnt_freeres(nfsdclient2, (xdrproc_t)xdr_readres, (caddr_t)&result.readres);
      Dprintf(D_DCACHE, "got second part: data_len %d\n",
      		read_res.readres_u.reply.data.data_len);
		
      result.readres = read_res;            
      result.readres.readres_u.reply.data.data_val = &data_cache_buffer[0];
      if(memcpy(result.readres.readres_u.reply.data.data_val,
      		read_res.readres_u.reply.data.data_val,
		read_res.readres_u.reply.data.data_len) == NULL) {
	Dprintf(D_DCACHE, "error memory copy\n");
	free(read_res.readres_u.reply.data.data_val);
        goto no_caching;
      }
      free(read_res.readres_u.reply.data.data_val);
    } /* end if crossed boundary */  
      
    if((match1 == 1 && !crossed_boundary) || (match1 == 1 && match2 == 1)) {
#ifdef CALL_PROFILING
      log_cache_hits(rqstp->rq_proc); /*log cache hit */
#endif
    }
    if(match1 || crossed_boundary)
      data_attr_cache_hit=1;

    goto out;
  } /* endif  data_cache_enabled */

no_caching: /* dcache disabled or cache error */
  FORWARD_REQUEST(nfsproc_read_2, nfsdclient2, "nfsproc_read_2");  
  if (res->status!=NFS_OK) { 
    return res->status;
  }
/*
  if(res->readres_u.reply.data.data_len == 
  			result.readres.readres_u.reply.data.data_len) {
    if(memcmp(res->readres_u.reply.data.data_val,
    		result.readres.readres_u.reply.data.data_val,
		result.readres.readres_u.reply.data.data_len) != 0) {
      Dprintf(D_DCACHE, "data contents do not match\n");
    }
  }
  else Dprintf(D_DCACHE, "data lengths do not match\n");
  data_attr_cache_hit = 0;*/
  result.readres = *res;
  Dprintf(D_DCACHE, "forwarding ... returned data_len %d\n",
  	result.readres.readres_u.reply.data.data_len);

out:
  /* Map attributes and return to client */
  map_return_attributes(&result.readres.readres_u.reply.attributes, 
			nfsmount, rqstp);
  return (result.readres.status);
}

int
nfsd_nfsproc_writecache_2(void *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_writecache_2, nfsdclient2, "nfsproc_writecache_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* No need to map attributes */
  return (NFS_OK);
}

int
nfsd_nfsproc_write_2(writeargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  attrstat *res;
  
  Dprintf(D_DCACHE, "*** WRITE: offset %d totalcount %d \n",
    argp->offset, argp->totalcount);   
  hexdump(argp->file.data, NFS_FHSIZE, D_DCACHE);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries (both data and attribute) if necessary */
  if(data_cache_enabled) {
    dcache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_write_2, nfsdclient2, "nfsproc_write_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* Map attributes and return to client */
  result.attrstat = *res;
  map_return_attributes(&result.attrstat.attrstat_u.attributes,
			nfsmount, rqstp);
  return (result.attrstat.status);
}

int
nfsd_nfsproc_create_2(createargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  diropres *res;
  attrstat res_attr;
  int insres, lkupres;

  Dprintf(D_DCACHE, "*** CREATE: %s\n", argp->where.name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_create_2, nfsdclient2, "nfsproc_create_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  hexdump(res->diropres_u.diropres.file.data, NFS_FHSIZE, D_DCACHE);

  if(attr_cache_enabled) { /* insert created file handle into LOOKUP cache */
    insres = cache_insert(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)res);
    Dprintf(D_ACACHE, "cache insert: create_2(%i,%i) = %s\n",
	    rqstp->rq_vers, NFSPROC_LOOKUP, (insres==0 ? "OK":"ERROR"));

    /* also insert created file's attributes into GETATTR cache */
    lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR, 
		   (void *)&res->diropres_u.diropres.file, (void *)&res_attr);
    Dprintf(D_DCACHE, "cache lookup: create_2(%i,%i) = %s\n",
	  rqstp->rq_vers, NFSPROC_GETATTR,
	  (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));	    

    res_attr.status = res->status;
    res_attr.attrstat_u.attributes = res->diropres_u.diropres.attributes;
    insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&res_attr);
    Dprintf(D_ACACHE, "cache insert: create_2(%i,%i) = %s\n",
	    rqstp->rq_vers, NFSPROC_GETATTR, (insres==0 ? "OK":"ERROR"));	    
  }

  /* Map attributes and return to client */
  result.diropres = *res;
  map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
  return (result.diropres.status);
}

int
nfsd_nfsproc_remove_2(diropargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  nfsstat *res;

  Dprintf(D_DCACHE, "**** REMOVE: %s\n", argp->name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_remove_2, nfsdclient2, "nfsproc_remove_2");
  if(*res != NFS_OK) { 
    return *res;
  }
  
  /* No need to map attributes */
  result.nfsstat = *res;
  return (result.nfsstat);
}

int
nfsd_nfsproc_rename_2(renameargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  nfsstat *res;

  Dprintf(D_DCACHE, "*** RENAME: %s -> %s\n", argp->from.name, argp->to.name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_rename_2, nfsdclient2, "nfsproc_rename_2");
  if(*res != NFS_OK) { 
    return *res;
  }
  
  /* No need to map attributes */
  result.nfsstat = *res;
  return (result.nfsstat);
}

int
nfsd_nfsproc_link_2(linkargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  nfsstat *res;

  Dprintf(D_DCACHE, "*** LINK: %s\n", argp->to.name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_link_2, nfsdclient2, "nfsproc_link_2");
  if(*res != NFS_OK) { 
    return *res;
  }
  
  /* No need to map attributes */
  result.nfsstat = *res;
  return (result.nfsstat);
}

int
nfsd_nfsproc_symlink_2(symlinkargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  nfsstat *res;

  Dprintf(D_DCACHE, "*** SYMLINK: from %s to %s\n", argp->from.name, argp->to);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_symlink_2, nfsdclient2, "nfsproc_symlink_2");
  if(*res != NFS_OK) { 
    return *res;
  }
  
  /* No need to map attributes */
  result.nfsstat = *res;
  return (result.nfsstat);
}

int
nfsd_nfsproc_mkdir_2(createargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  diropres *res;
  int insres, lkupres;
  attrstat res_attr;

  Dprintf(D_DCACHE, "*** MKDIR: %s\n", argp->where.name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_mkdir_2, nfsdclient2, "nfsproc_mkdir_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  hexdump(res->diropres_u.diropres.file.data, NFS_FHSIZE, D_DCACHE);
  
  if(attr_cache_enabled) { /* insert created file handle into LOOKUP cache */
    insres = cache_insert(rqstp->rq_vers, NFSPROC_LOOKUP, (void *)res);
    Dprintf(D_ACACHE, "cache insert: mkdir_2(%i,%i) = %s\n",
	    rqstp->rq_vers, NFSPROC_LOOKUP, (insres==0 ? "OK":"ERROR"));

    /* also insert created file's attributes into GETATTR cache */
    lkupres = cache_lookup(rqstp->rq_vers, NFSPROC_GETATTR, 
		   (void *)&res->diropres_u.diropres.file, (void *)&res_attr);
    Dprintf(D_DCACHE, "cache lookup: mkdir_2(%i,%i) = %s\n",
	  rqstp->rq_vers, NFSPROC_GETATTR,
	  (lkupres==1 ? "HIT" : (lkupres==0 ? "MISS":"ERROR")));	    

    res_attr.status = res->status;
    res_attr.attrstat_u.attributes = res->diropres_u.diropres.attributes;
    insres = cache_insert(rqstp->rq_vers, NFSPROC_GETATTR, (void *)&res_attr);
    Dprintf(D_ACACHE, "cache insert: create_2(%i,%i) = %s\n",
	    rqstp->rq_vers, NFSPROC_GETATTR, (insres==0 ? "OK":"ERROR"));	    
  }

  /* Map attributes and return to client */
  result.diropres = *res;
  map_return_attributes(&result.diropres.diropres_u.diropres.attributes,
			nfsmount, rqstp);
  return (result.diropres.status);
}

int
nfsd_nfsproc_rmdir_2(diropargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  nfsstat *res;

  Dprintf(D_DCACHE, "*** RMDIR: %s\n", argp->name);
  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Invalidate cache entries if necessary */
  if(attr_cache_enabled) {
    cache_inval(rqstp->rq_vers, rqstp->rq_proc, (void *)argp);
  }

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_rmdir_2, nfsdclient2, "nfsproc_rmdir_2");
  if(*res != NFS_OK) { 
    return *res;
  }
  
  /* No need to map attributes */
  result.nfsstat = *res;
  return (result.nfsstat);
}

int
nfsd_nfsproc_readdir_2(readdirargs *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  readdirres *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_readdir_2, nfsdclient2, "nfsproc_readdir_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* No need to map attributes */
  result.readdirres = *res;
  return (result.readdirres.status);
}

int
nfsd_nfsproc_statfs_2(nfs_fh *argp, struct svc_req *rqstp)
{
  nfsstat astat;
  statfsres *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST(nfsproc_statfs_2, nfsdclient2, "nfsproc_statfs_2");
  if (res->status!=NFS_OK) { 
    return res->status;
  }

  /* No need to map attributes */
  result.statfsres = *res;
  return (result.statfsres.status);
}

#ifdef NFSV3
int
nfsd_nfsproc_null_3(void *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  getattr3res *res; 

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_null_3, nfsdclient3, "nfsproc_null_3");

  /* No need to map attributes in NULL call */
  return (NFS_OK);
}

int
nfsd_nfsproc_getattr_3(nfs_fh3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  getattr3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_getattr_3, nfsdclient3, "nfsproc_getattr_3");

  /* Map attributes and return to client */
  result.getattr3res = *res;
  map_return_attributes3(&result.getattr3res.getattr3res_u.attributes,
			 nfsmount, rqstp);
  return (result.getattr3res.status);
}

int
nfsd_nfsproc_setattr_3(setattr3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  wccstat3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_setattr_3, nfsdclient3, "nfsproc_setattr_3");

  /* Map attributes and return to client */
  result.wccstat3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.\
			   after.post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.wccstat3.status);
}

int
nfsd_nfsproc_lookup_3(diropargs3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  lookup3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_lookup_3, nfsdclient3, "nfsproc_lookup_3");

  /* Map attributes and return to client */
  result.lookup3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.lookup3res.lookup3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.lookup3res.lookup3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.lookup3res.lookup3res_u.resok.dir_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.lookup3res.status);
}

int
nfsd_nfsproc_access_3(access3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  access3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_access_3, nfsdclient3, "nfsproc_access_3");

  /* Map attributes and return to client */
  result.access3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.access3res.access3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.access3res.access3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.access3res.status);
}

int
nfsd_nfsproc_readlink_3(nfs_fh3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  readlink3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_readlink_3, nfsdclient3, "nfsproc_readlink_3");

  /* Map attributes and return to client */
  result.readlink3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.readlink3res.readlink3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.readlink3res.readlink3res_u.resok.\
			 symlink_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.readlink3res.status);
}

int
nfsd_nfsproc_read_3(read3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  read3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_read_3, nfsdclient3, "nfsproc_read_3");

  /* Map attributes and return to client */
  result.read3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.read3res.read3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.read3res.read3res_u.resok.file_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.read3res.status);
}

int
nfsd_nfsproc_write_3(write3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  write3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_write_3, nfsdclient3, "nfsproc_write_3");

  /* Map attributes and return to client */
  result.write3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.write3res.write3res_u.resfail.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.write3res.write3res_u.resok.file_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.write3res.status);
}

int
nfsd_nfsproc_create_3(create3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  diropres3 *res;

   /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_create_3, nfsdclient3, "nfsproc_create_3");

  /* Map attributes and return to client */
  result.diropres3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.\
			 after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.diropres3.diropres3_u.resok.\
			 obj_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.diropres3.status);
}

int
nfsd_nfsproc_mkdir_3(mkdir3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  diropres3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_mkdir_3, nfsdclient3, "nfsproc_mkdir_3");

  /* Map attributes and return to client */
  result.diropres3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.diropres3.status);
}

int
nfsd_nfsproc_symlink_3(symlink3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  diropres3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_symlink_3, nfsdclient3, "nfsproc_symlink_3");

  /* Map attributes and return to client */
  result.diropres3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.diropres3.status);
}

int
nfsd_nfsproc_mknod_3(mknod3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  diropres3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_mknod_3, nfsdclient3, "nfsproc_mknod_3");

  /* Map attributes and return to client */
  result.diropres3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.diropres3.diropres3_u.resfail.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.diropres3.diropres3_u.resok.dir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.diropres3.diropres3_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.diropres3.status);
}

int
nfsd_nfsproc_remove_3(diropargs3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  wccstat3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_remove_3, nfsdclient3, "nfsproc_remove_3");

  /* Map attributes and return to client */
  result.wccstat3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.wccstat3.status);
}

int
nfsd_nfsproc_rmdir_3(diropargs3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  wccstat3 *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_rmdir_3, nfsdclient3, "nfsproc_rmdir_3");

  /* Map attributes and return to client */
  result.wccstat3 = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.wccstat3.wccstat3_u.wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.wccstat3.status);
}

int
nfsd_nfsproc_rename_3(rename3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  rename3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_rename_3, nfsdclient3, "nfsproc_rename_3");

  /* Map attributes and return to client */
  result.rename3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.rename3res.rename3res_u.res.\
			   fromdir_wcc.after.post_op_attr_u.attributes,
			   nfsmount, rqstp);
    map_return_attributes3(&result.rename3res.rename3res_u.res.\
			   todir_wcc.after.post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.rename3res.rename3res_u.res.\
			 fromdir_wcc.after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.rename3res.rename3res_u.res.\
			 todir_wcc.after.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.rename3res.status);
}

int
nfsd_nfsproc_link_3(link3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  link3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_link_3, nfsdclient3, "nfsproc_link_3");

  /* Map attributes and return to client */
  result.link3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.link3res.link3res_u.res.linkdir_wcc.after.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    map_return_attributes3(&result.link3res.link3res_u.res.file_attributes.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.link3res.link3res_u.res.linkdir_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  map_return_attributes3(&result.link3res.link3res_u.res.file_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.link3res.status);
}

int
nfsd_nfsproc_readdir_3(readdir3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  readdir3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_readdir_3, nfsdclient3, "nfsproc_readdir_3");

  /* Map attributes and return to client */
  result.readdir3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.readdir3res.readdir3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.readdir3res.readdir3res_u.resok.\
			 dir_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.readdir3res.status);
}

int
nfsd_nfsproc_readdirplus_3(readdirplus3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  readdirplus3res *res;
  struct entryplus3 *scan;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_readdirplus_3, nfsdclient3,
		   "nfsproc_readdirplus_3");

  /* Map attributes and return to client */
  result.readdirplus3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.readdirplus3res.readdirplus3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.readdirplus3res.readdirplus3res_u.resok.\
			 dir_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  /* need to map all returned attributes */
  for(scan = result.readdirplus3res.readdirplus3res_u.resok.reply.entries;
      scan != NULL; scan = scan->nextentry) {
    map_return_attributes3(&scan->name_attributes.post_op_attr_u.attributes,
			   nfsmount, rqstp);
  }
  return (result.readdirplus3res.status);
}

int
nfsd_nfsproc_fsstat_3(nfs_fh3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  fsstat3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_fsstat_3, nfsdclient3, "nfsproc_fstat_3");

  /* Map attributes and return to client */
  result.fsstat3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.fsstat3res.fsstat3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.fsstat3res.fsstat3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.fsstat3res.status);
}

int
nfsd_nfsproc_fsinfo_3(nfs_fh3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  fsinfo3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_fsinfo_3, nfsdclient3, "nfsproc_fsinfo_3");

  /* Map attributes and return to client */
  result.fsinfo3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.fsinfo3res.fsinfo3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.fsinfo3res.fsinfo3res_u.resok.obj_attributes.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.fsinfo3res.status);
}

int
nfsd_nfsproc_pathconf_3(nfs_fh3 *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  pathconf3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_pathconf_3, nfsdclient3, "nfsproc_pathconf_3");

  /* Map attributes and return to client */
  result.pathconf3res = *res;
  if (res->status != NFS3_OK) {
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.pathconf3res.pathconf3res_u.resfail.\
			   post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.pathconf3res.pathconf3res_u.resok.\
			 obj_attributes.post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.pathconf3res.status);
}

int
nfsd_nfsproc_commit_3(commit3args *argp, struct svc_req *rqstp)
{
  nfsstat3 astat;
  commit3res *res;

  /* Authenticate client */
  AUTHENTICATE_CLIENT;

  if(nfsmount->o.read_only) /* block writes to RO file systems */
    return NFSERR_ROFS;

  /* Forward request to server */
  FORWARD_REQUEST3(nfsproc3_commit_3, nfsdclient3, "nfsproc_commit_3");

  /* Map attributes and return to client */
  result.commit3res = *res;
  if (res->status != NFS3_OK) { 
    /* map uid/gid even if this call failed (resfail) */
    map_return_attributes3(&result.commit3res.commit3res_u.resfail.\
			   after.post_op_attr_u.attributes,
			   nfsmount, rqstp);
    return(res->status);
  }
  map_return_attributes3(&result.commit3res.commit3res_u.resok.file_wcc.after.\
			 post_op_attr_u.attributes,
			 nfsmount, rqstp);
  return (result.commit3res.status);
}
#endif /* NFSV3 */



/****************************************************************************
 * MAIN function
 *  read command line arguments; initialize signal handlers;
 *  setup NFS client and server
 ****************************************************************************/
int main(int argc, char **argv)
{
  char	*auth_file = NULL;
  char  *cachebasepath = NULL;
  char  *cookiefile = NULL;
  char  *cachesessionid = NULL;
  int   attrcachesize = DEFAULT_ATTR_CACHE_SIZE;
  int   attrcacheassoc = DEFAULT_ATTR_CACHE_ASSOC;
  int   attrcachebanks = DEFAULT_ATTR_CACHE_BANKS;
  int   datacachesize = DEFAULT_DATA_CACHE_SIZE;
  int   datacacheassoc = DEFAULT_DATA_CACHE_ASSOC;
  int   datacachebanks = DEFAULT_DATA_CACHE_BANKS;
  int	foreground = 0;
  int	nfsportin = 0;
  int   nfsportout = 0;
  int   rootsock = 0;
  int   transport = NFS_OVER_TCP;
  int   c;
  char  localhostname[HOSTNAMESIZE];

  program_name = argv[0];
  localhostname[0] = '\x0';
  chdir("/");

  /* Parse the command line options and arguments. */
  opterr = 0;
  while ((c = getopt_long(argc, argv, shortopts, longopts, NULL)) != EOF)
    switch (c) {
    case 'h':
      usage(stdout, 0);
      break;
    case 'd':
      enable_logging(optarg);
      break;
    case 'F':
      foreground = 1;
      break;
    case 'f':
      auth_file = optarg;
      break;
    case 'i':
      cachesessionid = optarg;
      break;
    case 'C':
#ifndef COOKIE_AUTH
      fprintf(stderr, "Cookie authentication has not been"
	      " enabled in this executable.\n");
      exit(-1);
#endif /* COOKIE_AUTH */
      cookiefile = optarg;
      break;
    case 'V':
      nfsvers = atoi(optarg);
      if(nfsvers<2 || nfsvers>3) {
	fprintf(stderr, "NFS version %i not supported\n",nfsvers);
	exit(-1);
      }
#ifndef NFSV3
      if(nfsvers==3) {
	fprintf(stderr, "NFS version %i not supported\n",nfsvers);
	exit(-1);
      }
#endif /* NFSV3 */
      break;
    case 'A':
      attr_cache_enabled = 1;
      break;
    case 'D':
      data_cache_enabled = 1;
      break;
    case 'c':
      cachebasepath = optarg;
      break;
    case 'z':
      attrcachesize = atoi(optarg);
      break;
    case 'a':
      attrcacheassoc = atoi(optarg);
      break;
    case 'b':
      attrcachebanks = atoi(optarg);
      break;
    case 'e':
      datacachesize = atoi(optarg);
      break;
    case 'o':
      datacacheassoc = atoi(optarg);
      break;
    case 'k':
      datacachebanks = atoi(optarg);
      break;
    case 'H':
      strncpy(&localhostname[0], optarg, HOSTNAMESIZE);
      break;
    case 'P':
      nfsportin = atoi(optarg);
      if (nfsportin < 0) {
	fprintf(stderr, "nfsd: bad port number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 'O':
      nfsportout = atoi(optarg);
      if (nfsportout <= 0) {
	fprintf(stderr, "nfsd: bad port number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 'S':
      rootsock = atoi(optarg);
      if (rootsock <= 0) {
	fprintf(stderr, "nfsd: bad socket number: %s\n",
		optarg);
	usage(stderr, -1);
      }
      break;
    case 't':
      if(strcmp(optarg,"tcp")==0) {
	transport = NFS_OVER_TCP;
      } else {
	if(strcmp(optarg,"udp")==0) {
	  transport = NFS_OVER_UDP;
	} else {
	  fprintf(stderr, "nfsd: bad transport: %s\n",
		  optarg);
	  usage(stderr, -1);
	}
      }
      break;
    case 'T':      
      sync_interval = atoi(optarg);
      if (sync_interval <= 0) {
	fprintf(stderr, "nfsd: bad sync interval value: %s\n",
		optarg);
	usage(stderr, -1);
      }
      sync_enabled = 1;
      break;
    case 's':
      if((send_sk_path = optarg) == NULL) {
	fprintf(stderr,"Must specify path to the session key\n");
	usage(stderr, -1);
      }
      send_sk_enabled = 1;
      break;
    case 'm':
      if((match_sk_path = optarg) == NULL) {
	fprintf(stderr,"Must specify path to the session key\n");
	usage(stderr, -1);
      } 
      match_sk_enabled = 1;
      break;
    case 'p':
      promiscuous = 1;
      break;
    case 'v':
      printf("%s\n", version);
      exit(0);
    case 0:
      break;
    case '?':
    default:
      usage(stderr, -1);
    } /* endof switch & while*/

  /* No more arguments allowed. */
  if (optind != argc)
    usage(stderr, -1);
  
  if(auth_file == NULL) {
    fprintf(stderr,"Must specify exports file\n");
    exit(-1);
  }

  /* needed for auth_clnt.c */
  authvers = nfsvers;

  /* Initialize logging. */
  log_open("nfsd", foreground);
  
  /* Initialize RPC stuff */
  rpc_init("nfsd", NFS_PROGRAM, nfsvers, nfs_dispatch,
	   nfsportin, NFS_MAXDATA);

  /* Obtain local host name for default forwarding address if
     not given as command-line argument (-H) */
  if(localhostname[0] == '\x0') { 
    if(gethostname(localhostname,HOSTNAMESIZE)!=0) {
      fprintf(stderr,"gethostname(): couldn't obtain local host name\n");
      exit(-1);
    }
  }

  /* initialize v2 client */
  if(nfsvers == 2) {
    if((nfsdclient2=open_nfsdclient(localhostname,nfsportout,2,rootsock,
				    transport))==NULL) {
      fprintf(stderr,"Couldn't create v2 client to %s:%i; bailing out\n",
	      localhostname,nfsportout);
      exit(-1);
    }
    authclient2 = nfsdclient2;
  }
#ifdef NFSV31
  /* initialize v3 client */
  if(nfsvers == 3) {
    if((nfsdclient3=open_nfsdclient(localhostname,nfsportout,3,rootsock,
				    transport))==NULL) {
      fprintf(stderr,"Couldn't create v3 client to %s:%i; bailing out\n",
	      localhostname,nfsportout);
      exit(-1);
    }
    authclient3 = nfsdclient3;
  }
#endif /* NFSV3 */

  authclient1 = NULL;

  /* Initialize attribute cache */
  if(attr_cache_enabled) {
    if((init_attr_cache(attrcachesize, attrcacheassoc, attrcachebanks,
			cachebasepath,cachesessionid))==-1) {
      Dprintf(L_ERROR, "nfsd: cannot initialize attribute cache\n");
      exit(-1);
    } else {
      umask(2); /* set mask for sharing cache banks among group */
    }
  }

  /* Initialize data cache */
  if(data_cache_enabled) {
    if((init_data_cache(datacachesize, datacacheassoc, datacachebanks,
			cachebasepath,cachesessionid))==-1) {
      fprintf(stderr, "nfsd: cannot initialize data cache\n");
      exit(-1);
    } else {
      umask(2); /* set mask for sharing cache banks among group */
    }
  }

  /* If not in foreground, fork off a child. */
  if (!foreground) {
    if ((c = fork()) > 0)
      /* parent exits */
      exit(0);
    if (c < 0) {
      Dprintf(L_FATAL, "nfsd: cannot fork: %s\n",
	      strerror(errno));
    }
  }

  /* Initialize the AUTH module. */
  auth_init(auth_file);

  if(send_sk_enabled || match_sk_enabled)
    if(sk_init(send_sk_path, match_sk_path) != 0) {
      Dprintf(L_ERROR, "Session key initialization error\n");
      exit(-1);
  }

  if (!foreground) {
    /* No more logging to stderr */
    background_logging();

    /* Now we remove ourselves from the foreground. */
    close(0);
    close(1);
    close(2);
#ifdef HAVE_SETSID
    setsid();
#else
    {
      int fd;
      if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
	ioctl(fd, TIOCNOTTY, (char *) NULL);
	close(fd);
      }
    }
#endif
  }


  /* Install signal handlers:
     USR1 toggles logging on/off
     USR2 dumps statistics to /tmp/gvfs.nfsd_profile.PID 
     TERM also dumps statistics, but then exits 
     QUIT flushes caches and dumps statistics
     HUP re-reads input files (exports, map) */
  install_signal_handler(SIGUSR1, toggle_logging);
  install_signal_handler(SIGUSR2,  sigusr2);
#ifdef CALL_PROFILING
  install_signal_handler(SIGQUIT,  flush_caches);
  gettimeofday (&inittime, NULL);
#endif
  install_signal_handler(SIGHUP,  reinitialize);
  install_signal_handler(SIGTERM, sigterm);
  /* atexit registers "terminate" to be called at normal
     program termination */
  atexit(terminate);
  
  if(attr_cache_enabled) {
    install_signal_handler(SIGALRM, setalarm);
    if(sync_enabled) alarm(sync_interval);
    Dprintf(D_DCACHE, "setup alarm(%d)\n", sync_interval);
  }
    
  /* Run the NFS server. */
  fprintf(stdout,"Proxy started successfully.\n");
  fflush(stdout);
  svc_run();

  Dprintf(L_ERROR, "Oh no Mr. Bill... nfs_server() returned!\n");
  exit(-1);
}


/****************************************************************************
 * Shows program usage
 ****************************************************************************/
static void
usage(FILE *fp, int n)
{
  fprintf(fp,
	  "\nUsage: %s -f exports-file\n"
	  "       [-d kind] [-H servername] [-S socketd] [-t tcp|udp]\n"
          "       [-C cookiefile] [-P inputport] [-O outputport]\n"
	  "       [-V nfsvers] [-F] [-h] [-p] [-v] [-A] [-D]\n"
          "       [-T syncinterval] [-i cachesessionid] [-c cachebasepath]\n"
          "       [-z attrcachesize] [-a attrcacheassoc] [-b attrcachebanks] \n"
	  "       [-e datacachesize] [-o datacacheassoc] [-k datacachebanks] \n"
	  "       [-s sendskpath] [-m matchskpath] \n"
	  , program_name);
  fprintf(fp,
          "\nOptions: \n"
	  "       -d: enables debugging mode (auth,all,call,devtab,acache,dcache,general,rmtab,stale,ugid)\n"
	  "       -F: puts process to foreground\n"
	  "       -h: prints help\n"
	  "       -H: specifies server name (local hostname is used if not specified)\n"
	  "       -S: specifies socket\n"
	  "       -t: specifies transport protocol (tcp <default> or udp)\n"
	  "       -V: specifies NFS version (2 <default> or 3)\n"
	  "       -C: specifies cookie file\n"
	  "       -A: enables attributes cache (only works with NFS version 2)\n"
	  "       -D: enables data cache (only works with NFS version 2)\n"
	  "       -T: specifies periodical sync interval (seconds) of caches with server\n"
	  "       -i: specifies cache session id (required if cache is enabled)\n"
	  "       -c: specifies cache base path (required if cache is enabled)\n"
	  "       -z: specifies attributes cache size (default is 4096 blocks)\n"
	  "       -a: specifies attributes cache associativity (default is 8)\n"
	  "       -b: specifies attributes cache bank number (default is 64)\n"
	  "       -e: specifies data cache size (default is 65536 blocks)\n"
	  "       -o: specifies data cache associativity (default is 16)\n"
	  "       -k: specifies data cache bank number (default is 128)\n"
	  "       -p: enables promiscuous mode\n"
	  "       -v: prints version number\n"
	  "       -f: specifies path of exports file (required)\n"
	  "       -P: specifies input port number (a random port is chosen and printed out if not specified)\n"
	  "       -O: specifies output port number (a random port is chosen and printed out if not specified)\n"
	  "       -s: specifies path of the session key to be sent\n"
	  "       -m: specifies path of the session key to match with\n");
  fprintf(fp,
          "\nSignals:\n"
	  "       -USR1: toggles logging on/off\n"
	  "       -USR2: dumps statistics to /tmp/gvfs.nfsd profile.PID\n"
	  "       -HUP:  re-reads input files (exports and map)\n"
	  "       -ALRM: syncs caches with server\n"
	  "       -QUIT: flushes caches and dumps statistics to /tmp/gvfs.nfsd-profile.PID\n"
	  "       -TERM: exits and dumps statistics to /tmp/gvfs.nfsd-profile.PID\n");

  exit(n);
}


/****************************************************************************
 * Signal handler for SIGTERM
 ****************************************************************************/
static RETSIGTYPE
sigterm(int sig)
{
  /* Dump statistics if enabled */
#ifdef CALL_PROFILING
  dump_stats(SIGTERM);
#endif /* CALL_PROFILING */
  exit(0);
}

/****************************************************************************
 * Signal handler for SIGUSR2
 ****************************************************************************/
static RETSIGTYPE
sigusr2(int sig)
{
  /* Dump statistics if enabled */
#ifdef CALL_PROFILING
  dump_stats(SIGUSR2);
#endif /* CALL_PROFILING */
}

/**************************************************************************** 
 * Exit subroutine called upon normal program termination 
 ****************************************************************************/
static void
terminate(void)
{
  /* finish server */
  rpc_exit(NFS_PROGRAM, nfsvers);
  /* finish client(s) */
  close_nfsdclient(nfsdclient2);
#ifdef NFSV3
  close_nfsdclient(nfsdclient3);
#endif /* NFSV3 */
}

/****************************************************************************
 * Signal handler for SIGQUIT
 ****************************************************************************/
RETSIGTYPE flush_caches(sig)
{
  static volatile int	inprogress = 0;
  
  /* in the middle of processing a request; do not reinitialize yet */
  if (_rpcsvcdirty) {
    need_flush = 1;
    return;
  }
  if (inprogress++)	/* SIGUSR2 handling in progress already? */
    return;
  /* not in the middle of processing a request, not handling another
     SIGUSR - go ahead and flush caches */
  if(attr_cache_enabled)
    flush_attr_cache();	/* attribute cache */
  if(data_cache_enabled)
    flush_data_cache();   /* data cache */
  dump_stats(0);         /* dump profiling stats */
  inprogress = 0;
  need_flush = 0;
}

/****************************************************************************
 * Signal handler for SIGHUP 
 ****************************************************************************/
RETSIGTYPE reinitialize(sig)
{
  static volatile int	inprogress = 0;
  
  /* in the middle of processing a request; do not reinitialize yet */
  if (_rpcsvcdirty) {
    need_reinit = 1;
    return;
  }
  if (inprogress++)	/* SIGHUP handling in progress already? */
    return;
  /* not in the middle of processing a request, not handling another
     SIGHUP - go ahead and reinitialize input files */
  auth_init(NULL);	/* auth_init remembers the exports file name */
  if(match_sk_enabled)
    if(sk_init(send_sk_path, match_sk_path) != 0) {
      Dprintf(L_ERROR, "Session key initialization error\n");
      match_sk_enabled = 0;
  }
  inprogress = 0;
  need_reinit = 0;
}

/****************************************************************************
 * Signal handler for SIGALRM
 ****************************************************************************/
RETSIGTYPE setalarm(sig)
{
  static volatile int	inprogress = 0;
  
  /* in the middle of processing a request; do not reinitialize yet */
  if (_rpcsvcdirty) {
    need_alarm = 1;
    return;
  }
  if (inprogress++)	/* SIGALRM handling in progress already? */
    return;    
  
  /* not in the middle of processing a request, not handling another
     SIGALRM - go ahead and set alarm */ 
  if(attr_cache_enabled) flush_attr_cache();
  
  if(sync_enabled && attr_cache_enabled) alarm(sync_interval);
   
  inprogress = 0;
  need_alarm = 0;
}
